#!/usr/bin/python

import sys
import threading
import traceback

from SerialConnector.Crc import Crc


from SmartMeshSDK.ApiException import ConnectionError, \
                                      CommandError

try:
    import serial
except ImportError:
    output  = ''
    output += 'Could not load the serial module.\n'
    output += 'Please install PySerial from http://pyserial.sourceforge.net/,'
    output += 'then run this script again.\n'
    raw_input(output)
    sys.exit()

import logging
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('Hdlc')
log.setLevel(logging.ERROR)
log.addHandler(NullHandler())

class Hdlc(threading.Thread):
    
    _BAUDRATE      = 115200 # the default baudrate used by this module
    
    _HDLC_FLAG     = 0x7e   # the HDLC flag at the beginning/end of a frame
    _HDLC_ESCAPE   = 0x7d   # escape character when the HDLC flag in payload
    _HDLC_MASK     = 0x20   # mask used to recover from escape characters
    
    _FCS_LENGTH    = 2      # number of bytes in the FCS field
    
    def __init__(self,rxcallback,connectcallback):
        
        # log
        log.info('Creating object')
        
        # store params
        self.rxcallback           = rxcallback
        self.connectcallback      = connectcallback
        
        # initialize parent class
        threading.Thread.__init__(self)
        self.name                 = 'HDLC'
        self.daemon               = True
        
        # local variables
        self.crc                  = Crc()
        self.connected            = False
        self.comPort              = ''
        self.pyserialHandler      = ''
        self.busySending          = threading.Lock()
        self.busyReceiving        = False
        self.lastRxByte           = self._HDLC_FLAG
        
        # initialize state
        self._restart()
    
    def run(self):
        
        # log
        log.info('thread started')
        
        try:
        
            while self.connected==True:
                try:
                    
                    # received a byte
                    try:
                        rxByte = self.pyserialHandler.read(1)
                    except Exception as err:
                        # work-around for bug in pyserial
                        # https://sourceforge.net/tracker/?func=detail&aid=3591432&group_id=46487&atid=446302
                        raise serial.SerialException(str(err))
                    if not len(rxByte):
                        raise serial.SerialException()
                    
                    # convert the received byte from a character to an int
                    rxByte = ord(rxByte)
                    if      (
                                (not self.busyReceiving)             and
                                self.lastRxByte==self._HDLC_FLAG     and
                                rxByte!=self._HDLC_FLAG
                            ):
                        # start of frame
                        
                        self.busyReceiving            = True
                        self.receivedFrame            = {}
                        self.receivedFrame['payload'] = [rxByte]
                    
                    elif    (
                                self.busyReceiving                   and
                                rxByte!=self._HDLC_FLAG
                            ):
                        # middle of frame
                        
                        if rxByte==self._HDLC_ESCAPE:
                            self._escape = True
                            # do not add this byte to the payload
                        else:
                            if self._escape == True:
                                self.receivedFrame['payload'].append(rxByte^self._HDLC_MASK)
                                self._escape = False
                            else:
                                self.receivedFrame['payload'].append(rxByte)
                    
                    elif    (
                                self.busyReceiving                   and
                                rxByte==self._HDLC_FLAG
                            ):
                        # end of frame
                        
                        self.busyReceiving       = False
                        
                        # split payload and fcs
                        if len(self.receivedFrame['payload'])>self._FCS_LENGTH:
                            temp_payload                   = self.receivedFrame['payload']
                            self.receivedFrame['payload']  = temp_payload[:-2]
                            self.receivedFrame['fcs']      = temp_payload[-2:]
                            
                            # check fcs, write 'valid' field
                            recalculatedCrc                = self.crc.calculate(self.receivedFrame['payload'])
                            if recalculatedCrc==self.receivedFrame['fcs']:
                                self.receivedFrame['valid'] = True
                            else:
                                self.receivedFrame['valid'] = False
                            
                            # log
                            if log.isEnabledFor(logging.DEBUG):
                                output     = []
                                output    += ['\nreceivedFrame:']
                                output    += self._formatFrame(self.receivedFrame)
                                log.debug('\n'.join(output))
                            
                            # callback
                            if self.receivedFrame['valid']==True:
                                try:
                                    self.rxcallback(self.receivedFrame['payload'])
                                except (ConnectionError,CommandError) as err:
                                    output = "@Hdlc: {0}".format(err)
                                    log.error(output)
                                    print (output)
                        else:
                            output = "@Hdlc: received hdlc frame too short"
                            log.error(output)
                        self._restart()
                    
                    # remember the last byte I received
                    self.lastRxByte = rxByte
                    
                except serial.SerialException:
                    self.connected = False
        
        except Exception as err:
            output  = []
            output += ['===== crash in thread {0} ====='.format(self.name)]
            output += ['\nerror:\n']
            output += [str(err)]
            output += ['\ncall stack:\n']
            output += [traceback.format_exc()]
            output  = '\n'.join(output)
            print (output) # critical error
            log.critical(output)
            raise
        
        del self.crc
        del self.pyserialHandler
        self.connectcallback(self.connected)
        
        # log
        log.info('thread ended')
    
    #======================== public ==========================================
    
    def connect(self,comPort,baudrate=_BAUDRATE):
        self.comPort         = comPort
        try:
            self.pyserialHandler = serial.Serial(self.comPort,baudrate=baudrate)
            self.pyserialHandler.setRTS(False)
            self.pyserialHandler.setDTR(True)
        except serial.serialutil.SerialException as err:
            output = "could not open {0}@{1}baud, reason: {2}".format(self.comPort,baudrate,err)
            log.warning(output)
            raise ConnectionError(output)
        else:
            log.info("opened port {0}@{1}baud".format(self.comPort,baudrate))
            self._restart()
            self.connected = True
            self.connectcallback(self.connected)
            self.name      = '{0}_HDLC'.format(self.comPort)
            self.start()
        return self
    
    def send(self,message):
        
        if not self.connected:
            output = 'not connected'
            log.error(output)
            raise ConnectionError(output)
        
        # calculate fcs
        packetToSend              = {}
        packetToSend['payload']   = message
        packetToSend['fcs']       = self.crc.calculate(packetToSend['payload'])
        packetToSend['valid']     = True
        
        # assemble packet
        packetBytes               = packetToSend['payload']+packetToSend['fcs']
        
        # add HDLC escape characters
        index = 0
        while index<len(packetBytes):
            if packetBytes[index]==self._HDLC_FLAG or packetBytes[index]==self._HDLC_ESCAPE:
                packetBytes.insert(index,self._HDLC_ESCAPE)
                index += 1
                packetBytes[index] = packetBytes[index]^self._HDLC_MASK
            index += 1
        
        # add HDLC flags
        packetBytes.insert(0,self._HDLC_FLAG)
        packetBytes.insert(len(packetBytes),self._HDLC_FLAG)
        
        # log
        if log.isEnabledFor(logging.DEBUG):
            output           = []
            output          += ['\npacketToSend:']
            output          += self._formatFrame(packetToSend)
            log.debug('\n'.join(output))
        
        # reconstitute byteArray
        byteArray = bytes(packetBytes)

        # send over serial port
        try:        
            with self.busySending:
                numWritten   = self.pyserialHandler.write(byteArray)
        except IOError as e:
            raise ConnectionError(str(e))

        if numWritten!=len(packetBytes):
            output = 'wrote '+str(numWritten)+' bytes, expected '+str(len(packetBytes))
            log.error(output)
            raise ConnectionError(output)
    
    def disconnect(self):
        log.info("disconnect")
        if self.connected==True:
            self.pyserialHandler.close()
        self.comPort         = ''
        self.name            = 'HDLC'
    
    #======================== support methods for Python "with" statement =====
    
    def __enter__(self):
        pass
    
    def __exit__(self, type, value, tb):
        self.disconnect()

    #======================== private =========================================
    
    def _restart(self):
        self._escape         = False
        self.busyReceiving   = False
        self.lastRxByte      = self._HDLC_FLAG
    
    def _formatFrame(self,frame):
        returnVal  = []
        returnVal += [' - payload: {0}'.format(' '.join(["%.02x"%b for b in frame['payload']]))]
        returnVal += [' - fcs:     {0}'.format(' '.join(["%.02x"%b for b in frame['fcs']]))]
        returnVal += [' - valid:   {0}'.format(frame['valid'])]
        return returnVal
    


"""
This module implements the Gateway States Machine Mixin class.

This separates the boilerplate declaration of the states machine
from the Gateway processing code.
"""

# ============================ imports =========================================
import transitions
from typing import List
from txp.devices.gateway_global_state import GatewayGlobalState
from txp.common.configuration import GatewayConfig
from txp.common.edge import EdgeDescriptor, MachineMetadata
from txp.devices.gateway_states_enums import *
from pathlib import Path
import shelve
import os
from txp.common.config import settings
import logging
import datetime

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Definition of the Gateway States Machine class.
# ==============================================================================
class GatewayStatesMachineMixin:
    """The Gateway States Machine Mixin offers the functionality of the State
    machine to the Gateway class.

    This class takes care of all the boilerplate of the transitions.Machine
    declaration.

    This class holds a reference to the GatewayGlobalState dataclass which contains
    the Gateway state to be collected in production, and to be persisted.

    Transitions callbacks defined for the States Machine are methods defined by
    the Gateway.
    """

    def __init__(self):
        # State Machine initialization
        # The machine starts in Uninitialized. Later on, when everything is ready,
        # The Gateway will call the first real state.
        self._states_machine: transitions.Machine = transitions.Machine(
            model=self,
            states=GatewayStates,
            initial=GatewayStates.Uninitialized,
            before_state_change="_update_visited_states",
        )

        self._visited_states_stack: List[str] = []
        """_states_registry: A stack of visited states. Currently being used in unit testing to check
        visited states"""

        self._init_transitions()

        self.global_state: GatewayGlobalState = None
    # ================ Property definitions for syntactic sugar access ================================
    @property
    def gateway_config(self) -> GatewayConfig:
        return self.global_state.gateway_conf

    @gateway_config.setter
    def gateway_config(self, new_gateway_conf: GatewayConfig) -> None:
        self.global_state.gateway_conf = new_gateway_conf

    @property
    def machines_configurations(self) -> List[MachineMetadata]:
        return self.global_state.machines_configurations

    @machines_configurations.setter
    def machines_configurations(self, machines_configs: List[MachineMetadata]) -> None:
        self.global_state.machines_configurations = machines_configs

    @property
    def edges_configurations(self) -> List[EdgeDescriptor]:
        return self.global_state.edges_configurations

    @edges_configurations.setter
    def edges_configurations(self, edges_configs: List[EdgeDescriptor]) -> None:
        self.global_state.edges_configurations = edges_configs

    @property
    def status(self):
        """Returns the status value form the Global State.
        Status value is an enumerated member from one of the states status enumerations"""
        return self.global_state.status

    @status.setter
    def status(self, new_status):
        self.global_state.status = new_status

    # ========================== Class Private Methods ================================
    def _load_global_state(self) -> None:
        """Loads the appropriate global state for the Gateway object.

        New Gateway in Production -> New GatewayGlobalState instance.
        Produced Gateway ready to Start -> Recovered GatewayGlobalState from disk.

        If a Gateway started the production phase, and never completed it for some reason,
        then the Gateway will transition to an Error State. For that the marker files are used.
        """
        if not os.path.exists(settings.gateway.initialized_marker_file):
            log.info('No previous marker files found. Initialize a new Gateway')
            self.global_state = self._get_new_state()

        else: 
            # The Gateway must have finished the production phase. Otherwise, Error.
            if not os.path.exists(settings.gateway.ready_marker_file):
                # log.info("The Gateway never completed the Production Phase. Transition to Error State.")
                # self.global_state = self._get_new_state()  # Assigns a not None state
                # self._production_phase_not_completed()  # Transitions to error
                log.info('Previous marker files found, but production never completed. Initialize a new Gateway')
                self.global_state = self._get_new_state()

            else:
                global_state_recovered_db = shelve.open(
                    settings.gateway.global_state_persistence_path
                )
                global_state_recovered = global_state_recovered_db["global_state"]
                global_state_recovered_db.close()
                self.global_state = global_state_recovered
                self.global_state.boot_time = datetime.datetime.now().replace(microsecond=0)

    def _get_new_state(self) -> GatewayGlobalState:
        """Returns a instance of a new Global State with default values"""
        return GatewayGlobalState(
                GatewayConfig(
                    "",
                    "",
                    "",
                    settings.gateway.cloud_region,
                    [],
                    ""
                ),
                [],
                [],
                GatewayStates.Startup,
                StartupStatus.RequestRequiredValues,
                0,
                {},
                0,
                datetime.datetime.now().replace(microsecond=0),
                "0"
        )

    def _init_transitions(self) -> None:
        """Add the transitions to the initialized gateway states machine

        Each transition can have different types of callbacks that will be
        executed when the transition is called.
        https://github.com/pytransitions/transitions#callback-execution-order

        Since the Gateway holds the states machine as internal attribute, the
        transitions will invoke callbacks that are methods of the Gateway.
        """

        transitions_list = [
            {
                "trigger": "begin_startup",
                "source": GatewayStates.Uninitialized,
                "dest": GatewayStates.Startup,
                "after": [
                    "_update_global_state_value",
                    "_step_into_startup"
                ],
            },
            {
                "trigger": "_startup_failed",
                "source": GatewayStates.Startup,
                "dest": GatewayStates.StartupError,
                "after": ["_update_global_state_value", "_step_into_startup_error"],
            },
            {
                "trigger": "_startup_completed",
                "source": GatewayStates.Startup,
                "dest": GatewayStates.ConfigurationReception,
                "after": [
                    "_update_global_state_value",
                    "_step_into_configuration_reception",
                ],
            },
            {
                # This transition can be called from the Startup State OR
                # the Uninitialized State when recovering from previous completed Startup State
                "trigger": "_configuration_reception_completed",
                "source": [
                    GatewayStates.ConfigurationReception,
                    GatewayStates.Uninitialized,
                    GatewayStates.Ready
                ],
                "dest": GatewayStates.ConfigurationValidation,
                "after": [
                    "_update_global_state_value",
                    "_step_into_configuration_validation",
                ],
            },
            {
                "trigger": "_configuration_validation_completed",
                "source": [GatewayStates.ConfigurationValidation, GatewayStates.Uninitialized],
                "dest": GatewayStates.DevicesPairing,
                "after": [
                    "_update_global_state_value",
                    "_step_into_devices_pairing",
                ]
            },
            {
                "trigger": "_production_phase_not_completed",
                "source": GatewayStates.Uninitialized,
                "dest": GatewayStates.Error,
                "after": [
                    "_update_global_state_value",
                    "_step_into_production_not_completed_error"
                ]
            },
            {
                # This transition will put the Gateway in the PreparingDevices state.
                # If a Gateway is already 'produced', it can jump right away to this state
                # to prepare all the DriversHost and Drivers.
                "trigger": "_device_pairing_completed",
                "source": [GatewayStates.DevicesPairing, GatewayStates.ConfigurationValidation],
                "dest": GatewayStates.PreparingDevices,
                "after": [
                    "_update_global_state_value",
                    "_step_into_preparing_devices"
                ]
            },
            {
               # This transition will put the Gateway in the PreparingDevices state when
               # a new configuration has been received and parsed while the Gateway was in
               # Ready or Sample state.
               "trigger": "_prepare_devices_for_new_config",
               "source": [GatewayStates.Ready, GatewayStates.Uninitialized],
               "dest": GatewayStates.PreparingDevices,
               "after": [
                   "_update_global_state_value",
                   "_step_into_preparing_devices"
               ]
            },
            {
                "trigger": "_preparing_devices_completed",
                "source": [GatewayStates.PreparingDevices, GatewayStates.ConfigurationValidation],
                "dest": GatewayStates.Ready, # Uninitialized while production process gets implemented
                "after": [
                    "_update_global_state_value",
                    "_step_into_ready_state"
                ]
            },
            {
                "trigger": "_start_sampling",
                "source": GatewayStates.Ready,
                "dest": GatewayStates.Sampling,
                "after": [
                    "_update_global_state_value",
                    "_step_into_sampling_state"
                ]
            },
            {
                "trigger": "_sampling_completed",
                "source": GatewayStates.Sampling,
                "dest": GatewayStates.Ready,
                "after": [
                    "_update_global_state_value",
                    "_step_into_ready_state"
                ]
            }
        ]

        self._states_machine.add_transitions(transitions_list)

    def _update_visited_states(self, *args, **kwargs):
        """Updates the visited states stack"""
        self._visited_states_stack.insert(0, self.state)

    def _update_global_state_value(self) -> None:
        """Updates the state variable inside the GlobalState attribute to the state
        value inside the machine object"""
        self.global_state.state = self.state

    def _save_global_state(self) -> None:
        """Save the GlobalState instance attribute on disk using shelve."""
        log.debug("Trying to save Global State in file system")
        with shelve.open(settings.gateway.global_state_persistence_path) as db:
            db["global_state"] = self.global_state
            db.close()
        log.debug("Global State saved in file system.")

    def _create_initialized_marker_file(self) -> None:
        """Creates the marker file to indicate that the Gateway started a production phase.

        This marker file indicates that the Gateway started a production phase.
        """
        if not Path(settings.gateway.initialized_marker_file).exists():
            Path(settings.gateway.initialized_marker_file).touch()

    def _create_ready_marker_file(self) -> None:
        """Creates the marker file to indicate that the Gateway completed the production phase.
        """
        if not Path(settings.gateway.ready_marker_file).exists():
            Path(settings.gateway.ready_marker_file).touch(exist_ok=True)

    def _last_visited_state(self) -> GatewayStates:
        return self._visited_states_stack[0]




# Warning: Do NOT edit this file directly. Your changes may be overwritten. 
# This file is automatically generated by GenIpMgrSubscribe.py

import threading

from   SmartMeshSDK import ApiException

class IpMgrSubscribe(object):
    '''
    \brief Notification listener for IpMgrConnectorMux object
    '''
    
    class SubscribeError(Exception) :
        def __init__(self, msg) :
            self.msg = msg
        def __str__(self):
            return self.msg
    
    ERROR                = "error"
    FINISH               = "finish"
    NOTIFEVENT           = "notifEvent"
    NOTIFLOG             = "notifLog"
    NOTIFDATA            = "notifData"
    NOTIFIPDATA          = "notifIpData"
    NOTIFHEALTHREPORT    = "notifHealthReport"
    NOTIFRADIOTESTSTATSPER = "notifRadiotestStatsPER"
    ALLNOTIF             = [NOTIFEVENT, NOTIFLOG, NOTIFDATA, NOTIFIPDATA, NOTIFHEALTHREPORT, NOTIFRADIOTESTSTATSPER]
    EVENTMOTERESET       = "eventMoteReset"
    EVENTNETWORKRESET    = "eventNetworkReset"
    EVENTCOMMANDFINISHED = "eventCommandFinished"
    EVENTMOTEJOIN        = "eventMoteJoin"
    EVENTMOTEOPERATIONAL = "eventMoteOperational"
    EVENTMOTELOST        = "eventMoteLost"
    EVENTNETWORKTIME     = "eventNetworkTime"
    EVENTPINGRESPONSE    = "eventPingResponse"
    EVENTPATHCREATE      = "eventPathCreate"
    EVENTPATHDELETE      = "eventPathDelete"
    EVENTPACKETSENT      = "eventPacketSent"
    EVENTMOTECREATE      = "eventMoteCreate"
    EVENTMOTEDELETE      = "eventMoteDelete"
    EVENTJOINFAILED      = "eventJoinFailed"
    EVENTINVALIDMIC      = "eventInvalidMIC"

    _trNotifNameTable = {
    "eventMoteReset" : "notifEvent",
    "eventNetworkReset" : "notifEvent",
    "eventCommandFinished" : "notifEvent",
    "eventMoteJoin" : "notifEvent",
    "eventMoteOperational" : "notifEvent",
    "eventMoteLost" : "notifEvent",
    "eventNetworkTime" : "notifEvent",
    "eventPingResponse" : "notifEvent",
    "eventPathCreate" : "notifEvent",
    "eventPathDelete" : "notifEvent",
    "eventPacketSent" : "notifEvent",
    "eventMoteCreate" : "notifEvent",
    "eventMoteDelete" : "notifEvent",
    "eventJoinFailed" : "notifEvent",
    "eventInvalidMIC" : "notifEvent",

    }
    
    #======================== public ==========================================
    
    def __init__(self, ipMgrConnector) :
        # Structure of self._callback :
        #     Notification Name : 
        #         [0] - subscription mask mask, 
        #         [1] - cb-function. Notification is subscribed if [1]!=None, 
        #         [2] - transport for notification: True - reliable, false - unreliable
        self._callback = {
            self.ERROR             : [0x00, None, True],
            self.FINISH            : [0x00, None, True],
            self.NOTIFEVENT        : [0x02, None, True],
            self.NOTIFLOG          : [0x04, None, True],
            self.NOTIFDATA         : [0x10, None, True],
            self.NOTIFIPDATA       : [0x20, None, True],
            self.NOTIFHEALTHREPORT : [0x40, None, True],
            self.NOTIFRADIOTESTSTATSPER : [0x80, None, True],
        
        }
        self._con    = ipMgrConnector
        self._thread = None
        self._mask = self._unrlblMask = 0
        self._isStarted = False
        self._lock = threading.Lock()
        
    def start(self):
        '''
        \brief Start the subscriber _thread.
        '''
        
        if self._thread :   # Wait finish disconnect process
            try :
                self._thread.join(1.0)
                if  self._thread.isAlive() :
                    raise ApiException.ConnectionError("Already connected")
            except RuntimeError :
                pass    # Ignore join error
            self._thread = None 
        
        # Clear _callback table
        for i in self._callback :
            self._callback[i][1] = None
            self._callback[i][2] = True
        self._mask = self._unrlblMask = 0
        self._thread = threading.Thread(target = self._process) 
        self._thread.name = "IpMgrSubscribe"
        self._thread.start()
        self._isStarted = True
        
    def subscribe(self, notifTypes, fun, isRlbl):
        '''
        \brief Subscribe to notification(s).
        
        Calling this function multiple times will not cancel the effects of
        the previous calls.
        
        \pre Call start() before calling this function.
        
        \param notifTypes Type(s) of notification(s) to subscribe to. This can
            be a single string (when subscribing to a single notification), or
            a list of strings (when subscribing to multiple notifications).
            The list of possible types is:
            ERROR, FINISH, NOTIFEVENT, NOTIFLOG, NOTIFDATA, NOTIFIPDATA, NOTIFHEALTHREPORT, NOTIFRADIOTESTSTATSPER, ALLNOTIF

        \param fun The function to call when any of the notification types
            specified in the notifTypes parameter occurs. If you wish to assign
            a different _callback function to different notification types,
            call this function multiple times. The signature of the function
            needs to be fun(<notification name>, <notification parameter>),
            as described below.
        \param isRlbl define type of transport using for delivery 
             notification: reliable (True) or best effort (False)
        The _callback function is called with a notification name and a
        notification parameter. Depending on the type of notification, the
        parameter will be of a different format, according to the table below.
        
        <table>
            <tr><th>Notification Name   </th><th>Parameter</th>
            <tr><td>ERROR               </td><td>Exception</td>
            <tr><td>FINISH              </td><td>''</td>
            <tr><td>NOTIFLOG            </td><td>Tuple_notifLog</td>
            <tr><td>NOTIFDATA           </td><td>Tuple_notifData</td>
            <tr><td>NOTIFIPDATA         </td><td>Tuple_notifIpData</td>
            <tr><td>NOTIFHEALTHREPORT   </td><td>Tuple_notifHealthReport</td>
            <tr><td>NOTIFRADIOTESTSTATSPER</td><td>Tuple_notifRadiotestStatsPER</td>
            <tr><td>EVENTMOTERESET      </td><td>Tuple_eventMoteReset</td>
            <tr><td>EVENTNETWORKRESET   </td><td>Tuple_eventNetworkReset</td>
            <tr><td>EVENTCOMMANDFINISHED</td><td>Tuple_eventCommandFinished</td>
            <tr><td>EVENTMOTEJOIN       </td><td>Tuple_eventMoteJoin</td>
            <tr><td>EVENTMOTEOPERATIONAL</td><td>Tuple_eventMoteOperational</td>
            <tr><td>EVENTMOTELOST       </td><td>Tuple_eventMoteLost</td>
            <tr><td>EVENTNETWORKTIME    </td><td>Tuple_eventNetworkTime</td>
            <tr><td>EVENTPINGRESPONSE   </td><td>Tuple_eventPingResponse</td>
            <tr><td>EVENTPATHCREATE     </td><td>Tuple_eventPathCreate</td>
            <tr><td>EVENTPATHDELETE     </td><td>Tuple_eventPathDelete</td>
            <tr><td>EVENTPACKETSENT     </td><td>Tuple_eventPacketSent</td>
            <tr><td>EVENTMOTECREATE     </td><td>Tuple_eventMoteCreate</td>
            <tr><td>EVENTMOTEDELETE     </td><td>Tuple_eventMoteDelete</td>
            <tr><td>EVENTJOINFAILED     </td><td>Tuple_eventJoinFailed</td>
            <tr><td>EVENTINVALIDMIC     </td><td>Tuple_eventInvalidMIC</td>
        </table>
        
        \exception IpMgrSubscribe.SubscribeError The subscriber hasn't been
            started, or the notification type(s) specified is (are) not valid.
        '''
        
        if not self._isStarted :
            raise self.SubscribeError("Error: subscriber is not started")
        if isinstance(notifTypes, str) :
            notifTypes = [notifTypes]
        for nType in notifTypes :  # subscribe type validation
            if nType not in self._callback :
                raise self.SubscribeError("Error subscribe type: {0}".format(nType))
        
        self._lock.acquire()
        for nType in notifTypes :
            self._callback[nType][1] = fun
            self._callback[nType][2] = isRlbl
        self._lock.release()
        
        mask = unrlblMask = 0
        # Structure of self._callback.values() :
        #     [0] - subscription mask mask, 
        #     [1] - cb-function. Notification is subscribed if [1]!=None, 
        #     [2] - transport for notification: True - reliable, false - unreliable
        for cb in self._callback.values() :
            if cb[1] :
                mask = mask | cb[0]
            if cb[2] == False :
                unrlblMask = unrlblMask | cb[0] 
        if mask != self._mask or unrlblMask != self._unrlblMask :
            self._mask = mask
            self._unrlblMask = unrlblMask
            self._con.dn_subscribe([0,self._mask], [0,self._unrlblMask])

    #======================== private =========================================
    
    def _process(self):
        while True :
            try :
                notif = self._con.getNotification()
                name = notif[0]
                if name in self._trNotifNameTable :
                    name = self._trNotifNameTable[name]
                self._processOneNotif(name, notif[0], notif[1])
            except ApiException.QueueError:
                self._processOneNotif(self.FINISH, self.FINISH, '')
                self._isStarted = False
                break
            except Exception as ex :
                self._processOneNotif(self.ERROR, self.ERROR, ex)
    
    def _processOneNotif(self, notifType, notifName, payload):
        cb = self._getCallback(notifType)
        if cb : 
            cb(notifName, payload)
    
    def _getCallback(self, name) :
        res = None

        self._lock.acquire()
        if name in self._callback :
            res = self._callback[name][1]
        self._lock.release()
        
        return res



from txp.web.core_components.main_view import MainView
from txp.web.core_components.app_profiles import AppProfile
from txp.web.core_components.app_component import Card
import txp.cloud.controllers.firestore as firestore_controller
import txp.cloud.controllers.iot_core as iot_controller
from txp.cloud.common.utils import pull_current_configuration_from_firestore
import logging
from typing import Dict
import streamlit as st
import datetime
import pytz
import pandas as pd

from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class DeviceHealthView(MainView):
    """The Device Health View"""

    def __init__(self):
        super(DeviceHealthView, self).__init__(component_key="device_health_view")
        self.gateways_data = None
        self.edges_data = None
        self.config: None
        self.machine_per_edges = None
        self.jobs_data = None
        self.gateway_to_render = None
        self.current_edges = None
        self._data_built = False

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.RESUMEN

    @classmethod
    def get_display_name(cls) -> str:
        return "Salud de Dispositivo"

    def _build_submenu_state(self) -> Dict:
        # Hardcoded state for demonstration
        pass

    def _build_content_state(self) -> Dict:
        pass

    def _update_edges(self):
        selected_edges = st.session_state[f"{self.__class__.__name__}_edge_multibox"]
        self.current_edges = selected_edges

    def _update_gateway(self):
        gateway = st.session_state[f"{self.__class__.__name__}_gateway_selectbox"]
        self.gateway_to_render = gateway
        self.current_edges = [
            edge["logical_id"] for edge in self.mapping_edges_that_send_data_to_gateway()
        ]

    def mapping_edges_that_send_data_to_gateway(self):
        current_edges = []
        for machine in self.gateways_data[self.gateway_to_render]["machines_ids"]:
            data = self.app_state.project_model.machines_table[machine].reprJSON()
            for edge in data['edges_ids']:
                edge_dict = self.app_state.project_model.edges_table[edge].reprJSON()
                if edge_dict not in current_edges:
                    current_edges.append(edge_dict)
        return current_edges

    def build_state(self):
        if not self._data_built:
            db = firestore_controller.get_authenticated_client(
                self.app_state.authentication_info.role_service_account_credentials
            )
            self.config = pull_current_configuration_from_firestore(db, self._tenant_id).to_dict()
            db.close()

            self.gateways_data = {
                key: self.app_state.project_model.gateways_table[key].reprJSON() for key in self.app_state.project_model.gateways_table.keys()
            }

            self.edges_data = {key: self.app_state.project_model.edges_table[key].reprJSON() for key in self.app_state.project_model.edges_table.keys()}

            self.jobs_data = {
                key: self.app_state.project_model.jobs_table_by_gateway[key].reprJSON()
                for key in self.app_state.project_model.gateways_table.keys()
            }

            self.machine_per_edges = self.get_machine_to_edge()

            self.gateway_to_render = next(iter(self.gateways_data))

            self.current_edges = [
                 edge["logical_id"] for edge in self.mapping_edges_that_send_data_to_gateway()
            ]

            self.iot = iot_controller.get_authenticate_client(
                 self.app_state.authentication_info.role_service_account_credentials
            )

            self.device_state = {key: iot_controller.get_device_state(
                 self.iot, self.gateways_data[key]['project_id'],
                 self.gateways_data[key]['cloud_region'],
                 self.gateways_data[key]['registry_id'],
                 self.gateways_data[key]['gateway_id']) for key in self.app_state.project_model.gateways_table.keys()}

            self._data_built = True

    def change_utc_timezone(self, timestamp):
        utc = pytz.timezone('UTC')
        timezone = pytz.timezone("America/Mexico_City")
        date_time = pd.to_datetime(timestamp)
        localized_timestamp = utc.localize(date_time)
        new_timezone = localized_timestamp.astimezone(timezone)
        strtime = new_timezone.strftime("%m/%d/%Y, %H:%M:%S")
        return strtime

    def _render_submenu(self) -> None:

        st.selectbox(
            label="Gateways de este proyecto",
            options=self.gateways_data.keys(),
            key=f"{self.__class__.__name__}_gateway_selectbox",
            on_change=self._update_gateway,
        )

        label = f"Dispositivos vinculados a {self.gateway_to_render}"

        edges = [
            edge["logical_id"] for edge in self.mapping_edges_that_send_data_to_gateway()
        ]

        st.multiselect(
            label,
            options=edges,
            key=f"{self.__class__.__name__}_edge_multibox",
            on_change=self._update_edges,
            default=self.current_edges,
        )
        st.markdown("* * *")
        st.markdown("**<p align='center'>Estado general</p>**", unsafe_allow_html=True)

        tasks = self.jobs_data[self.gateway_to_render]["tasks"]
        st.write(f"📥 Paquetes recibidos: **50**")
        st.write(f"🕚 Paquetes en cola: **10**")

        st.write(f"🔗 Total de dispositivos vinculados: **{len(self.current_edges)}**")

        st.write(
            f"📡 Total de activos monitoreados: **{len(self.gateways_data[self.gateway_to_render]['machines_ids'])}**"
        )

    def last_gateway_task(self):
        tz = pytz.timezone("America/Mexico_City")
        tasks = self.jobs_data[self.gateway_to_render]["tasks"]
        task_info = {}
        for i, task in enumerate(tasks):
            start_date = datetime.datetime.strptime(
                    task["parameters"]["start_date"], "%Y-%m-%d"
                ).date()
            start_time= datetime.datetime.strptime(
                    task["parameters"]["start_time"], "%H:%M"
                ).time()
            end_date = datetime.datetime.strptime(
                task["parameters"]["end_date"], "%Y-%m-%d"
            ).date()
            end_time = datetime.datetime.strptime(
                task["parameters"]["end_time"], "%H:%M"
            ).time()
            full_start_datetime = tz.localize(datetime.datetime.combine(start_date,start_time))
            full_end_datetime = tz.localize(datetime.datetime.combine(end_date, end_time))
            task_info['start_datetime'] = full_start_datetime.strftime("%Y-%m-%d %H:%M:%S")
            task_info['end_datetime'] = full_end_datetime.strftime("%Y-%m-%d %H:%M:%S")
            task_info['observation_time'] = task['parameters']['sampling_window'].observation_time
            task_info['sampling_time'] = task['parameters']['sampling_window'].sampling_time
            task_info['machines'] = task['machines']
        return task_info

    def get_last_actualization_gateway(self):
        conected_edges = []
        for edge,state in self.app_state.project_model.edges_states_table.items():
            if edge in self.current_edges:
                if not state is None:
                    conected_edges.append(state)

        if any(d["driver_state"] == "Disconnected" for d in conected_edges):
            st.warning("Estado: se registraron dispositivos desconectados")
        # TODO computing the Gateway State should be a logic on its own, and should consider all edge cases
        # elif any("state" not in d and not d.get('type', None) =='ACTUATOR_DEVICE' for d in conected_edges):
        #     st.warning("Estado: se registraron dispositivos sin información de estado")
        else:
            st.success("Estado: CONECTADO")

        cards=[]
        if self.device_state[self.gateway_to_render]:
            heartbeat = self.change_utc_timezone(self.device_state[self.gateway_to_render].heartbeat)
            last_state_time = self.change_utc_timezone(self.device_state[self.gateway_to_render].last_state_time)
            last_config_ack = self.change_utc_timezone(self.device_state[self.gateway_to_render].last_config_ack)
            last_error_time = self.change_utc_timezone(self.device_state[self.gateway_to_render].last_error_time)


            last_actualization_card = Card(
                card_row=1,
                card_column_start=1,
                card_column_end=3,
                card_subheader="Última actividad",
                content={"Señal de monitoreo": [heartbeat],
                         "Último estado del gateway": [last_state_time],
                         "Envío de configuración": [last_config_ack],
                         "Versión de la configuración": [self.config["configuration_id"]],
                         "Último error recibido": [last_error_time],
                         "Mensaje de error": [self.device_state[self.gateway_to_render].last_error_msg],
                         },
                labels=[
                    "Señal de monitoreo",
                    "Último estado del gateway",
                    "Envío de configuración",
                    "Versión de la configuración",
                    "Último error recibido",
                    "Mensaje de error",
                ],
                vertical_style=True,
            )
            cards.append(last_actualization_card)
        else:
            st.info('Este dispositivo no está conectado a IoT')

        last_task_info = self.last_gateway_task()

        tasks_of_gateway_card = Card(
                card_row=2,
                card_column_start=1,
                card_column_end=3,
                card_subheader="Configuración de la última tarea",
                content=last_task_info,
                labels=[
                    "Fecha de inicio",
                    "Fecha de finalización",
                    "Tiempo de observación",
                    "Tiempo de muestreo",
                    "Máquinas"
                ],
                vertical_style=False,
        )
        cards.append(tasks_of_gateway_card)

        self.render_card(
            cards
        )

        st.text("\n")
        return None

    def get_machine_to_edge(self):
        machines_with_edges = {
            machine: self.app_state.project_model.machines_table[machine].edges_ids
            for machine in self.app_state.project_model.machines_table.keys()
        }
        edges_and_monitored_machines = {}
        for motor, edges in machines_with_edges.items():
            for edge in edges:
                edges_and_monitored_machines.setdefault(edge, []).append(motor)

        return edges_and_monitored_machines


    def get_last_actualization_edge(self, edge):
        st.markdown(f'### **Dispositivo**: {edge["logical_id"]}.')
        with st.expander("Estado", expanded=True):
            state_dict = self.app_state.project_model.edges_states_table[edge['logical_id']]
            if state_dict:
                if "driver_state" in state_dict:
                    if state_dict["driver_state"] == "Connected":
                        st.success("DriverState.CONECTADO")
                        st.caption(f'Desde: {state_dict["since_datetime"]}')
                    elif state_dict["driver_state"] == "Sampling":
                        st.success("DriverState.RECOLECTANDO DATOS")
                        st.caption(f'Desde: {state_dict["since_datetime"]}')
                    elif edge.get('type', 'ACTUATOR_DEVICE'):
                        st.success("DriverState.CONECTADO")
                    else:
                        st.warning("Sin información del estado")
            else:
                st.error("DriverState.DESCONECTADO")
            st.write(f'**Tipo de dispositivo**: {edge["device_kind"]}')
            st.write(
                f'**Última recepción de evento de telemetría**: {self.config["since"]}'
            )
            st.write(f"**Magnitudes de percepción:**")
            for perception in edge["perceptions"]:
                st.write(f"{perception}")

            monitored_machines = []
            for m in self.gateways_data[self.gateway_to_render]["machines_ids"]:
                monitored_machines.append(m)

            for machine in monitored_machines:
                if machine in self.machine_per_edges[edge["logical_id"]]:
                    id = f"**Activo**: {machine}"
                    if self.app_state.project_model.machines_table[machine].model != 'No Aplica':
                        st.caption(
                            f"{id}, modelo {self.app_state.project_model.machines_table[machine].model}"
                        )
                    else:
                        st.caption(
                            f"{id}"
                        )

    def _render_content(self) -> None:

        self.build_state()

        st.header("Salud de los Dispositivos")

        st.markdown(f"#### **Gateway**: {self.gateway_to_render}.")

        self.get_last_actualization_gateway()
        st.markdown("* * *")

        for edge in self.current_edges:
            self.get_last_actualization_edge(self.edges_data[edge])



import threading
import socket
import select
import struct

from IpMgrConnectorMux.MuxMsg import MuxMsg

from   SmartMeshSDK import ApiException,                   \
                           ApiConnector
from   SmartMeshSDK.ApiDefinition import IpMgrDefinition

class IpMgrConnectorMuxInternal(ApiConnector.ApiConnector ) :
    '''
    \ingroup ApiConnector
    
    \brief Internal class for IP manager connector, through Serial Mux.
    
    Members of class
        acknowledgeBuf - binary payload of acknowledge
        ackCmdId       - command ID from acknowledge
        sendSemaphor   - semaphore to wait for acknowledgement of command (threading.Semaphore)
        inputThread    - thread for processing input packets (threading.Thread) 
        socket         - TCP socket for connection with Serial Mux   
    '''
    PARAM_HOST         = 'host'
    PARAM_PORT         = 'port'
    PARAM_ISSENDHELLO  = 'isSendHello'

    DEFAULT_PARAM_HOST = '127.0.0.1'
    DEFAULT_PARAM_PORT = 9900
    
    _RC_OK         = 0  
    _RC_TIMEOUT    = 5
    
    def __init__(self, maxQSize = 100) :
        ApiConnector.ApiConnector.__init__(self, maxQSize)
        self.acknowledgeBuf = None
        self.ackCmdId = -1
        self.sendSemaphor = threading.BoundedSemaphore(1)
        self.sendLock = threading.Lock()
        self.socket = None
        self.inputThread = None
        self.muxMsg = MuxMsg.MuxMsg(self.processCmd)
        self.apiDef = IpMgrDefinition.IpMgrDefinition()
        self.notifIds = self.apiDef.getIds(self.apiDef.NOTIFICATION) 
        
    def connect(self, params = {}) :
        '''
        \brief Connect to device
        
        \param params Dictionary of connection parameters:
            - 'host' - IP address of Mux (default: '127.0.0.1')
            - 'port' - port of Mux (default: 9900)
            - 'isSendHello' - send Hello message after connection (default True)
        '''
        
        host = self.DEFAULT_PARAM_HOST 
        port = self.DEFAULT_PARAM_PORT
        isSendHello = True
        if self.PARAM_HOST in params and params[self.PARAM_HOST] :
            host = params[self.PARAM_HOST]
        if self.PARAM_PORT in params and params[self.PARAM_PORT] :
            port = int(params[self.PARAM_PORT])
        if self.PARAM_ISSENDHELLO in params :
            isSendHello = params[self.PARAM_ISSENDHELLO]
        
        if self.inputThread :   # Wait finish disconnect process
            try :
                self.inputThread.join(1.0)
                if  self.inputThread.isAlive() :
                    raise ApiException.ConnectionError("Already connected")
            except RuntimeError :
                pass    # Ignore join error
            self.inputThread = None 
        
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect( (host, port) )
            self.socket.setblocking(1)
        except socket.error as ex:
            raise ApiException.ConnectionError(str(ex))
        
        self.sendSemaphor.acquire(False)    # Clear semaphore
        # Start thread for processing input stream
        self.inputThread = threading.Thread(target = self.inputProcess)
        self.inputThread.name = "IpMgrConnectorMuxInternal"
        self.inputThread.start()
        ApiConnector.ApiConnector.connect(self)
        if isSendHello :
            self.sendHelloCmd()
        
    def disconnect(self, reason="") :
        if not self.isConnected :
            return
        try :
            self.socket.send("stop")
            self.socket.shutdown(socket.SHUT_RD)    # start disconnection
            self.socket.close()
        except socket.error :
            pass    # Ignore socket error 
        ApiConnector.ApiConnector.disconnect(self, reason)
        
    def send(self, cmdNames, params) :
        self.sendLock.acquire()
        try :
            if not self.isConnected :
                raise ApiException.ConnectionError("Disconnected")
    
            # Send data
            ApiConnector.log.debug("IO OUT.    {0} : {1}".format(cmdNames, params))
            (cmdId, paramsBinList) = self.apiDef.serialize(cmdNames, params)
            paramsBin = struct.pack('!'+str(len(paramsBinList))+'B', *paramsBinList) 
            ApiConnector.logDump(paramsBin, "RawIO OUT. Command ID: {0}".format(cmdId))
            packet = self.muxMsg.build_message(cmdId, paramsBin)
            self.acknowledgeBuf = None
            self.ackCmdId = -1
            try :
                self.socket.sendall(packet)
            except socket.error as way:
                # Socket error. Disconnect from device. Stop command processing
                reason = "IO output error [{0}] {1}".format(way.args[0], way.args[1])
                self.disconnect(reason)
                raise ApiException.ConnectionError(reason)
            
            # Waiting acknowledge
            self.sendSemaphor.acquire()
            if not self.isConnected :         # Disconnect happened during waiting ack.  
                raise ApiException.ConnectionError(self.disconnectReason)
                                                        
            # Process acknowledge
            cmdId = self.apiDef.nameToId(self.apiDef.COMMAND, (cmdNames[0],))
            if self.ackCmdId != cmdId :
                reason = "Unexpected acknowledge {0} for command {1} ({2})".format(self.ackCmdId, cmdId, cmdNames)
                self.disconnect(reason)
                raise ApiException.ConnectionError(reason)
    
            # Parse acknowledge
            ackList = struct.unpack('!'+str(len(self.acknowledgeBuf))+'B', self.acknowledgeBuf)
            (resCmdName, resParams) = self.apiDef.deserialize(self.apiDef.COMMAND, self.ackCmdId, ackList) 
            ApiConnector.log.debug("IO INP.    {0} : {1}".format(resCmdName, resParams))
            
            if self.apiDef.RC in resParams and resParams[self.apiDef.RC] != self._RC_OK : 
                if resParams[self.apiDef.RC] == self._RC_TIMEOUT :
                    raise ApiException.CommandTimeoutError(resCmdName)
                try:
                    desc = '({0})\n{1}'.format(
                        self.apiDef.responseFieldValueToDesc(
                            resCmdName,
                            self.apiDef.RC,
                            resParams[self.apiDef.RC],
                        ),
                        self.apiDef.rcToDescription(
                            resParams[self.apiDef.RC],
                            resCmdName,
                        ),
                    )
                except:
                    desc = None
                raise   ApiException.APIError(
                            cmd=resCmdName,
                            rc=resParams[self.apiDef.RC],
                            desc=desc
                        )
            
            self.ackCmdId = -1
            self.acknowledgeBuf = None
        finally:
            self.sendLock.release()
        return resParams
            
    def ackSignal(self):
        '''
        \brief Send signal 'Acknowledge received'
        '''
        try    : 
            self.sendSemaphor.release()
        except ValueError : 
            pass

    def inputProcess(self):
        '''
        \brief Processing device input 
        '''
        try :
            while True :
                select.select([self.socket], [], [self.socket])
                buf = self.socket.recv(4096)
                if not buf :
                    raise socket.error(0, "Connection close")
                self.muxMsg.parse(buf)
        except socket.error as way:
            # Disconnect process -------------------------------------------------
            if way.args[0] == 9 :   # 
                way = socket.error(0, "Connection close")
            ApiConnector.ApiConnector.disconnect(self, "Disconnect. Reason: {0} [{1}]".format(way.args[1], way.args[0]))
            self.acknowledgeBuf = None
            self.ackCmdId = -1
            self.ackSignal()
            try :
                self.socket.close()
            except socket.error :
                pass    # Ignore socket error 
    
    def processCmd(self, reserved, cmdId, payload):
        '''
        \brief deserialize and process command
        '''
        ApiConnector.logDump(payload, "RawIO INP. Command ID: {0}".format(cmdId))
        if cmdId in self.notifIds :
            try :
                payloadList = struct.unpack('!'+str(len(payload))+'B', payload)
                (notifNames, params) = self.apiDef.deserialize(self.apiDef.NOTIFICATION, cmdId, payloadList)
                ApiConnector.log.debug("IO INP.    {0} : {1}".format(notifNames, params))
                self.putNotification((notifNames, params))
            except ApiException.ConnectionError as ex:
                raise socket.error(0, ex.value)    # Initiate disconnection
            except Exception as ex :
                ApiConnector.log.error("Deserialization command {0}. Error {1}".format(cmdId, ex))
        else :
            self.ackCmdId = cmdId
            self.acknowledgeBuf = payload
            self.ackSignal()
    
    def sendHelloCmd(self):
        '''
        \brief Send Hello command
        '''
        res = self.send(["mux_hello"], {"version" : self.muxMsg.getVer(), "secret" :  self.muxMsg.getAuth()})
        return res



# ============================ imports =========================================
from txp.devices.gateway import Gateway
from txp.common.config import settings
import logging
import signal


log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# Transitions are verbose. Enable a lower level only if debugging the Gateway State
# Machine transitions.
logging.getLogger('transitions').setLevel(logging.WARNING)


_gateway: Gateway = None


def _graceful_quit(signal_received, frame):
    _gateway.stop()


# ==============================================================================
# Main function to be called from the CLI module.
# ==============================================================================
def main():
    """The main application flow for the first Beta release.

    Currently this main entrypoint is a hardcoded to ensure that
    the Rapsberry image is properly working.

    The expected behavior is to initialize the Gateway Process and create
    the processes for:
        - Hardcoded DriversHost (By hand)
        - The streamlit web UI. (By the gateway)
    """
    log.info("Start main application for the Gateway process")

    log.info("Configuring DriversHostManager class")

    log.info(f"Samples processing mode set to: {settings.gateway.streaming_mode}")

    log.info("Creating the Gateway object instance")
    gateway = Gateway()
    global _gateway
    _gateway = gateway

    signal.signal(signal.SIGINT, _graceful_quit)

    # Starts the Gateway. Currently this starts loads all the hardwired configurations in memory
    # TODO: We should catch global exceptions and runtime errors here, and clear resources.
    gateway.start()

    # Move the gateway to Ready State.
    # In Ready state the Gateway should wait for the configuration to arrive,
    #   because harcoded configurations were removed.
    # When production phase is done, this can be removed.
    # gateway._preparing_devices_completed()

    while True:
        pass



"""This module defines utilities used commonly in the TXP
packages when dealing with datetime module and configuration
values"""

from datetime import datetime, timedelta
from txp.common.configuration.job import SamplingParameters
from typing import Optional
from logging import Logger


def get_num_windows_in_task(
    task_parameters: SamplingParameters
):
    """Return the number of sampling windows that fit in the interval declared
    in the sampling parameters."""
    current_date = datetime.now().date()
    end_date = current_date
    if task_parameters.end_time <= task_parameters.start_time:  # ends next day.
        end_date = (datetime.now() + timedelta(days=1)).date()

    start_datetime = datetime.combine(current_date, task_parameters.start_time)
    end_datetime = datetime.combine(end_date, task_parameters.end_time)
    diff = end_datetime - start_datetime
    num_samplings = int(
        diff.total_seconds() / task_parameters.sampling_window.observation_time
    )

    return num_samplings



from communications import hdlc
from communications.iCOMOX_communication import class_iCOMOX_Communication
import communications.hdlc
import time
import common.helpers

class class_HDLC_Communication(class_iCOMOX_Communication):
    ACK_TIMEOUT         = 0.2   # 200 msec
    MAX_NO_ACK_RETRIES  = 3

    def __init__(self, comm_port, callback_process_message, callback_serial_comm_open=None, callback_serial_comm_close=None, callback_no_ack_timeout=None):  #   , no_msg_read_timeout_sec=3, callback_no_msg_read_timeout=None):
        class_iCOMOX_Communication.__init__(
            self=self,
            comm_port=comm_port,
            baudrate=115200,
            #duration_us_of_break_before_write=0,   # 0 is the default value
            callback_get_msg_size=None,
            callback_process_message=callback_process_message,
            callback_open=callback_serial_comm_open,
            callback_close=callback_serial_comm_close,
            no_msg_read_timeout_sec=0,
            callback_no_msg_read_timeout=None,
            serial_comm_thread=self.__serial_comm_thread__
        )
        self.callback_timeout = callback_no_ack_timeout
        self.no_ack_timeout = None
        self.no_ack_retries = None

    def __build_hdlc_frame__(self, msg):
        result = bytearray(b"\x7E")
        msg += hdlc.FCS16_Calculate(msg)
        for Ch in msg:
            if (Ch == 0x7D) or (Ch == 0x7E):
                result += bytearray(b"\x7D")
                result += bytearray([Ch ^ 0x20])
            else:
                result += bytearray([Ch])
        result += bytearray(b"\x7E")
        return result

    def __extract_msg_from_hdlc_frame__(self, hdlc_frame):
        msg = bytearray(b"")
        if (len(hdlc_frame) < 2) or (hdlc_frame[0] != 0x7E) or (hdlc_frame[-1] != 0x7E):
            return [bytearray(b""), False]
        escaped = False
        for Ch in hdlc_frame[1:-1]:
            if escaped:
                if (Ch == 0x5D) or (Ch == 0x5E):
                    msg += bytearray([Ch ^ 0x20])
                    escaped = False
                else:
                    return [bytearray(b""), False]
            else:
                if Ch == 0x7E:
                    return [bytearray(b""), False]
                elif Ch == 0x7D:
                    escaped = True
                else:
                    msg += bytearray([Ch])
        if len(msg) >= 3:    # at least one character to process + 2 bytes of FCS
            FCS = msg[-2:]
            msg = msg[:-2]
            if FCS != hdlc.FCS16_Calculate(msg):
                return [bytearray(b""), False]
            else:
                return [msg, True]
        else:
            return [bytearray(b""), False]

#    def bytearray_find(self, bytesarr, value, start=0):
#        for i in range(start, len(bytesarr)):
#            if bytesarr[i] == value:
#                return i
#        return -1

    def __extractable_size_of_hdlc_frame__(self, data):
        firstEscapeIndex = data.find(b"\x7E", 0)     #firstEscapeIndex = self.bytearray_find(bytesarr=data, value=0x7E, start=0)
        if firstEscapeIndex > 0:
            return firstEscapeIndex, False   # extract the leading firstEscapeIndex bytes, no HDLC frame was detected
        elif firstEscapeIndex < 0:
            return 0, False

        if len(data) < 2:
            return 0, False

        lastEscapeIndex = data.find(b"\x7E", 1)     #lastEscapeIndex = self.bytearray_find(bytesarr=data, value=0x7E, start=1)
        if lastEscapeIndex < 0:
            return 0, False

        return lastEscapeIndex - firstEscapeIndex + 1, True

    def write_hdlc_msg(self, msg):
        hdlc_frame = self.__build_hdlc_frame__(msg)
        # helpers.OUT("TX mesage: " + helpers.u8s_to_str(hdlc_frame, " ", ""))
        return self.write(hdlc_frame)

#    def read_hdlc_msg(self):
#        result = True
#        try:
#            hdlc_frame = self.serObj.read_until(terminator="\x7E")
#        except:
#            hdlc_frame = bytearray()
#            result = False
#        finally:
#            if len(hdlc_frame) == 0:
#                return hdlc_frame, result
#            else:
#                return self.__extract_msg_from_hdlc_frame__(hdlc_frame)

    # __serial_comm_thread__ is the thread function
    # It serves 2 purposes:
    # 1. Handle events of close by trying to reopen the port
    # 2. Read HDLC messages from the port and call OnProcessMessage() with the accumulated message
    def __serial_comm_thread__(self):
        if not callable(self.OnProcessMessages):
            raise Exception("communicationHDLC.__serial_comm_thread__: OnProcessMessages()"
                            " callback must be a callable function")
        # reopen loop
        while not self.Terminate:
            # Wait until communication is open
            while not self.Terminate and not self.is_open():
                time.sleep(0.5)

            # read loop
            received_bytes = bytearray()
            while not self.Terminate and self.is_open():
                try:
                    received_bytes += self.serObj.read(size=self.serObj.in_waiting)
                except:
                    received_bytes = bytearray()
                    self.close()
                    continue
                finally:
                    pass

                bytes_to_extract, hdlc_frame_detected = self.__extractable_size_of_hdlc_frame__(received_bytes)
                if hdlc_frame_detected:
                    hdlc_frame = received_bytes[:bytes_to_extract]
                    received_bytes = received_bytes[bytes_to_extract:]
                    received_msg, result = self.__extract_msg_from_hdlc_frame__(hdlc_frame=hdlc_frame)
                    if not result:   # if invalid message due to any reason: length, flag char in the middle, too short message, checksum failure
                        self.close()
                        break
                    elif len(received_msg) != 0:           # if valid message was received
                        # helpers.OUT("RX message: " + helpers.u8s_to_str(hdlc_frame, " ", ""))
                        if callable(self.OnProcessMessages):
                            self.OnProcessMessages(self, received_msg)
                else:
                    received_bytes = received_bytes[bytes_to_extract:]

        # Thread was requested to terminate, then close the serial port
        self.close()



#!/usr/bin/env python
import argparse
import logging
import os
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
import google.cloud.firestore as firestore
from typing import List
import json

from txp.cloud.pipelines import transformers as ts

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()
logger.setLevel(logging.INFO)

# Service account key path
# os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "../credentials/pub_sub_to_bigquery_credentials.json"
BIGQUERY_TABLE = "PROJECT_ID:DATASET_NAME.TABLE_NAME"
MODEL_SERVING_TOPIC_PREFIX = f'projects/{os.environ.get("GCP_PROJECT_ID", "tranxpert-mvp")}/topics/'
MODEL_SERVING_TOPIC_NAME = 'txp-model-serving-signals-test'
SIGNALS_QUEUE_COLLECTION = 'signals_queue_test'
DATA_LAKE = 'tranxpert-mvp-telemetry-data-lake-test'


def get_firestore_collection(signals_queue_collection) -> List[str]:
    signals = []
    db = firestore.Client()
    docs = db.collection(signals_queue_collection).stream()
    for doc in docs:
        doc_as_dictionary = doc.to_dict()

        if bool(doc_as_dictionary):
            signals.append(json.dumps(doc_as_dictionary))

    return signals


class BuildSignalsFromGcs(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, data_lake, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        import os
        from google.cloud import storage
        from txp.cloud.common import utils

        project_data_lake = data_lake
        project_name = os.environ.get("GCP_PROJECT_ID", "tranxpert-mvp")

        all_chunks = []
        storage_client = storage.Client()
        bucket = storage_client.bucket(project_data_lake, user_project=project_name)

        signal_json = json.loads(element)

        for k in signal_json:
            for path in signal_json[k]:
                blob = bucket.blob(path)
                all_chunks.append(json.loads(blob.download_as_string()))

        all_chunks.sort(key=lambda c: c["part_index"])

        yield utils.merge_signal_chunks(all_chunks)


class PopFirestoreQueue(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, signals_queue_collection, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        import google.cloud.firestore as firestore
        firestore_db = firestore.Client()
        signal_json = json.loads(element)
        path = signal_json["paths"][0]
        doc = path.split("/")[1]
        doc_id = doc.split("|")[:-2]
        doc_id = "|".join(doc_id)
        firestore_db.collection(signals_queue_collection).document(doc_id).delete()


class BatchPublishToPubSub(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, topic_path, **kwargs):
        from google.cloud import pubsub_v1
        publisher = pubsub_v1.PublisherClient()
        future = publisher.publish(topic_path, data=element)
        return future.result()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("--time_table", help="Output Time BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--fft_table", help="Output Fft BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--psd_table", help="Output Psd BigQuery Table", default=BIGQUERY_TABLE)

    parser.add_argument("--time_metrics_table", help="Output Time metrics BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--fft_metrics_table", help="Output Fft metrics BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--psd_metrics_table", help="Output Psd metrics BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--model_signals_topic_name", help="topic for processing signals with ml task",
                        default=MODEL_SERVING_TOPIC_NAME)
    parser.add_argument("--signals_queue_collection",
                        help="queue of signals stored in firestore, "
                             "this is where signals are going to be downloaded from",
                        default=SIGNALS_QUEUE_COLLECTION)
    parser.add_argument("--data_lake", help="data lake bucket name in gcp", default=DATA_LAKE)

    known_args, pipeline_args = parser.parse_known_args()
    model_serving_topic = MODEL_SERVING_TOPIC_PREFIX + known_args.model_signals_topic_name

    pipeline_options = PipelineOptions(pipeline_args)

    with beam.Pipeline(options=pipeline_options) as p:
        pending_signals = (
                p
                | "ReadFromFirestore" >> beam.Create(get_firestore_collection(known_args.signals_queue_collection))
        )

        (
                pending_signals
                | "DeleteFromFirestore" >> beam.ParDo(PopFirestoreQueue(), known_args.signals_queue_collection)
        )

        signal_collection = (
                pending_signals
                | "BuildSignalsFromGcs" >> beam.ParDo(BuildSignalsFromGcs(), known_args.data_lake)
        )

        ########################################################################################################################
        time_signal = (
                signal_collection
                | "TimeProcessing" >> beam.ParDo(ts.TimeProcessing())
        )
        (
                time_signal
                | "WriteTimeToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.time_table)
                | "WriteTimeToPubSub" >> beam.ParDo(BatchPublishToPubSub(), model_serving_topic)
        )
        (
                time_signal
                | "TimeMetrics" >> beam.ParDo(ts.TimeMetrics())
                | "WriteTimeMetricsToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.time_metrics_table)
                | "WriteTimeMetricsToPubSub" >> beam.ParDo(BatchPublishToPubSub(), model_serving_topic)
        )
        ########################################################################################################################
        fft_signal = (
                signal_collection
                | "FftProcessing" >> beam.ParDo(ts.FftProcessing())
        )
        (
                fft_signal
                | "WriteFftToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.fft_table)
                | "WriteFftToPubSub" >> beam.ParDo(BatchPublishToPubSub(), model_serving_topic)
        )
        (
                fft_signal
                | "FftMetrics" >> beam.ParDo(ts.FftMetrics())
                | "WriteFftMetricsToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.fft_metrics_table)
                | "WriteFftMetricsToPubSub" >> beam.ParDo(BatchPublishToPubSub(), model_serving_topic)
        )
        ########################################################################################################################
        psd_signal = (
                signal_collection
                | "PsdProcessing" >> beam.ParDo(ts.PsdProcessing())

        )
        (
                psd_signal
                | "WritePsdToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.psd_table)
                | "WritePsdToPubSub" >> beam.ParDo(BatchPublishToPubSub(), model_serving_topic)
        )
        (
                psd_signal
                | "PsdMetrics" >> beam.ParDo(ts.PsdMetrics())
                | "WritePsdMetricsToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.psd_metrics_table)
                | "WritePsdMetricsToPubSub" >> beam.ParDo(BatchPublishToPubSub(), model_serving_topic)
        )


if __name__ == "__main__":
    run()



"""Common classes for Edge devices support.

This module contains the following classes that are commonly used in
edge definition and usage:
    - EdgeType
    - MachineMetadata
    - EdgeDescriptor
"""

# ============================ imports =========================================
from enum import Enum
from dataclasses import dataclass
from typing import Dict, Optional, Tuple
from typing import List
import txp.common.protos.gateway_config_pb2 as gateway_config_proto
from txp.devices.drivers.usb.camera.camera_handler_base import CameraType
from google.protobuf import json_format
import dataclasses
import hashlib


class EdgeType(Enum):
    """Enumeration names for the different types of defined edge devices

    The purpose of this class is to provide typed constants for edge types.

    An ACTUATOR_DEVICE, is the type of device that can be use to multiply
        edges. Concretely, we currently support the Robotic arm which is
        the only supported ACTUATOR_DEVICE.
    """

    STREAM_ONLY_DEVICE = "STREAM_ONLY_DEVICE"
    ON_OFF_DEVICE = "ON_OFF_DEVICE"
    SMART_DEVICE = "SMART_DEVICE"
    ACTUATOR_DEVICE = "ACTUATOR_DEVICE",

    def __str__(self):
        return self.name

    @staticmethod
    def from_str(type_val: str):
        if type_val == str(EdgeType.STREAM_ONLY_DEVICE.name):
            return EdgeType.STREAM_ONLY_DEVICE
        elif type_val == str(EdgeType.ON_OFF_DEVICE):
            return EdgeType.ON_OFF_DEVICE
        elif type_val == str(EdgeType.SMART_DEVICE.name):
            return EdgeType.SMART_DEVICE
        elif type_val == str(EdgeType.ACTUATOR_DEVICE.name):
            return EdgeType.ACTUATOR_DEVICE
        else:
            raise NotImplementedError


class SingletonMeta(type):
    """
    This is meta class for implementation of a singleton.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


@dataclass
class MachineMetadata:
    """Encapsulates the metadata from the machine monitored by an edge
    device.

    TODO: This class should contain whatever informtion associated to
        the machine is required for collected data interpretation.

    TODO: this class should contain information about the bearing
        mechanism
    """

    machine_id: str
    model: str
    manufacturer: str
    edges_ids: List[str]
    state_manager: str = ""
    tasks: dict = dataclasses.field(default_factory=dict)

    """edges_ids: The list of edge ids associated with this machine"""

    def get_proto(self) -> gateway_config_proto.MachineMetadataProto:
        """Returns the proto for the object.

        Returns:
            A MachineMetadataProto instance from this object.
        """
        proto = gateway_config_proto.MachineMetadataProto(
            machine_id=self.machine_id,
            model=self.model,
            manufacturer=self.manufacturer
        )
        proto.edges_ids.extend(self.edges_ids)
        return proto

    @classmethod
    def build_from_dict(cls, d: Dict) -> Optional["MachineMetadata"]:
        optional_fields = ['state_manager',
                           'tasks']

        known_fields = list(map(
            lambda field: field.name in d if field.name not in optional_fields else True,
            dataclasses.fields(cls)
        ))
        # All the known fields are in the dict
        if not all(known_fields):
            return None

        # Remove unknown field from the dict
        known_fields = list(map(
            lambda field: field.name,
            dataclasses.fields(cls)
        ))

        invalid_keys = [key if key not in known_fields else None for key in d.keys()]

        new_dict = d.copy()
        for key in invalid_keys:
            if key:
                new_dict.pop(key)

        m_m = MachineMetadata(**new_dict)
        return m_m

    @staticmethod
    def build_from_proto(
        proto: gateway_config_proto.MachineMetadataProto,
    ) -> "MachineMetadata":
        """Returns the class instance from the proto value.

        Args:
            proto: The MachineMetadataProto to build the class instance from.

        Raises:
            ValueError: if the received proto is not an instance of MachineMetadataProto.
        """
        if isinstance(proto, gateway_config_proto.MachineMetadataProto):
            return MachineMetadata(
                proto.machine_id,
                proto.model,
                proto.manufacturer,
                list(proto.edges_ids)
            )

        raise ValueError(
            "Trying to build MachineMetadata from proto, but the proto instance is unknown"
        )

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON.

        The Dictionary will contain the values required to re-create the instance from the JSON
        object in a pythonic way:
            object = MyObject( **json.loads(json_string) )
        """
        return dict(machine_id=self.machine_id, model=self.model, manufacturer=self.manufacturer,
                    edges_ids=self.edges_ids, state_manager=self.state_manager, tasks=self.tasks)


@dataclasses.dataclass(frozen=True)
class VirtualSensorObject:
    """Class to track the necessary information required for a virtual sensor.

    For example, a Camera or Thermocamera configured in a robotic virtual edge.
    """
    physical_id: str = ""
    camera_type: CameraType = None

    def __bool__(self):
        return bool(str(self.physical_id))  # Empty physical ID evaluates to False

    def reprJSON(self) -> Dict:
        return dict(physical_id=self.physical_id, camera_type=str(self.camera_type))

    @classmethod
    def build_from_dict(cls, d: Dict) -> Optional["VirtualSensorObject"]:
        known_fields = list(map(
            lambda field: field.name in d,
            dataclasses.fields(cls)
        ))

        # All the known fields are in the dict
        if not all(known_fields):
            return None

        return cls(d['physical_id'], CameraType(d['camera_type']))


@dataclasses.dataclass(frozen=True)
class VirtualEdgeInformation:
    """Class to encapsulate the virtual information for an EdgeDescriptor.

    Args:
        position (Tuple): The captured position for this virtual edge.
        virtual_id_hash (str): The virtual ID produced for the captured positions.
            This virtual ID is used with the edge logical ID to get a new
            virtual logical ID.
    """
    position: Tuple
    virtual_id_hash: str

    def reprJSON(self) -> Dict:
        return dict(position=self.position, virtual_id_hash=self.virtual_id_hash)

    @classmethod
    def build_from_dict(cls, d: Dict) -> Optional["VirtualEdgeInformation"]:
        known_fields = list(map(
            lambda field: field.name in d,
            dataclasses.fields(cls)
        ))

        # All the known fields are in the dict
        if not all(known_fields):
            return None

        return cls(tuple(d["position"]), d["virtual_id_hash"])


@dataclass
class EdgeDescriptor:
    """This EdgeDescriptor encapsulates information about an edge.

    Specifically, and EdgeDescriptor holds the information to know:
        - What's the identity of the edge
        - What's the type of the edge
        - Which system driver interacts with the edge
        - Which machine is being monitored by that edge.
        - Configuration parameters for the edge.

    Args:
        logical_id: The serial defined in the IoT management as a registered device.
        device_kind: the device kind from the project model.
            For example: Voyager.
        device_type: the edge type from the project model.
            For example: SMART_DEVICE
        machine_metadata: The machine metadata for the machine monitored
            by the edge.
        edge_parameters: A dictionary of edge parameters.
        perceptions: A Dictionary containing the perceptions definition for the edge.
    """

    logical_id: str
    device_kind: str
    device_type: str
    edge_parameters: Dict
    perceptions: Dict

    def __post_init__(self):
        """Post initializations to keep consistency across the system.

        If the 'physical_key' is not in the edge_parameters, then the device hasn't been paired
        and a default empty value is set.
        """
        if "physical_id" not in self.edge_parameters:
            self.edge_parameters["physical_id"] = ""

    def get_proto(self) -> gateway_config_proto.EdgeDescriptorProto:
        """TODO: Gateway Packages send the EdgeDescriptorProto. That's not really required
            and the GatewayPackage could just remove the EdgeDescriptorProto dependency.
        """
        proto = gateway_config_proto.EdgeDescriptorProto(
            logical_id=self.logical_id,

            device_kind=str(self.device_kind),

            device_type=self.device_type,
        )
        # Edge parameters not required, and probably they will be deprecated as
        # as parameter of the EdgeDescriptorProto.
        # proto.edge_params.update(self.edge_parameters)
        return proto

    @staticmethod
    def build_from_proto(
        proto: gateway_config_proto.EdgeDescriptorProto,
    ) -> "EdgeDescriptor":
        """Returns the class instance from the proto value.

        Args:
            proto: The EdgeDescriptorProto to build the class instance from.

        Raises:
            ValueError: if the received proto is not an instance of EdgeDescriptorProto.

        TODO: Configuration objects are handled using JSON entities. Protos for
            configuration objects are legacy and they should be removed.
        """
        if isinstance(proto, gateway_config_proto.EdgeDescriptorProto):
            return EdgeDescriptor(
                proto.logical_id,
                proto.device_kind,
                proto.device_type,
                json_format.MessageToDict(proto.edge_params),
                {}
            )

        raise ValueError(
            "Trying to build EdgeDescriptor from proto, but the proto instance is unknown"
        )

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON.

        The Dictionary will contain the values required to re-create the instance from the JSON
        object in a pythonic way
            object = MyObject( **json.loads(json_string) )
        """
        return dict(logical_id=self.logical_id, device_kind=self.device_kind,
                    device_type=self.device_type, edge_parameters=self.edge_parameters,
                    perceptions=self.perceptions)

    @classmethod
    def build_from_dict(cls, d: Dict) -> Optional["EdgeDescriptor"]:
        known_fields = list(map(
            lambda field: field.name in d,
            dataclasses.fields(cls)
        ))

        # All the known fields are in the dict
        if not all(known_fields):
            return None

        # Remove unknown field from the dict
        known_fields = list(map(
            lambda field: field.name,
            dataclasses.fields(cls)
        ))

        invalid_keys = [key if key not in known_fields else None for key in d.keys()]

        new_dict = d.copy()
        for key in invalid_keys:
            if key:
                new_dict.pop(key)

        if "sensors" in new_dict["edge_parameters"]:
            new_dict['edge_parameters']["sensors"] = list(map(
                lambda sensor_d: VirtualSensorObject.build_from_dict(sensor_d),
                new_dict['edge_parameters']["sensors"]
            ))

            if not any(new_dict['edge_parameters']["sensors"]):
                return None

        if "virtual_device" in new_dict["edge_parameters"]:
            new_dict["edge_parameters"]["virtual_device"] = VirtualEdgeInformation.build_from_dict(
                new_dict["edge_parameters"]["virtual_device"]
            )

            if not new_dict["edge_parameters"]["virtual_device"]:
                return None

        edge_descriptor = cls(**new_dict)
        return edge_descriptor

    def set_physical_id(self, physical_id: str) -> None:
        self.edge_parameters.update({'physical_id': physical_id})

    def is_paired(self) -> bool:
        """Returns True if the device is paired."""
        return bool(self.edge_parameters['physical_id'])  # Empty physical ID will return false

    def set_virtual_sensors(self, sensors: List[VirtualSensorObject]) -> None:
        """Adds the sensors object configured to this virtual Edge descriptor"""
        self.edge_parameters['sensors'] = sensors

    def get_virtual_sensors(self) -> Optional[List[VirtualSensorObject]]:
        """Returns the Virtual sensors of this virtual EdgeDescriptor.
        """
        return self.edge_parameters.get('sensors', None)

    def set_virtual_edge_information(self, virtual_info: VirtualEdgeInformation) -> None:
        """Set the virtual edge information configured for this virtual Edge"""
        self.edge_parameters["virtual_device"] = virtual_info

    def get_virtual_edge_information(self) -> Optional[VirtualEdgeInformation]:
        """Returns the virtual edge information saved in this virtual edge"""
        return self.edge_parameters.get("virtual_device", None)

    def is_actuator_device(self) -> bool:
        return EdgeType.from_str(self.device_type) == EdgeType.ACTUATOR_DEVICE

    @staticmethod
    def get_hashed_virtual_id(positions: Tuple) -> str:
        """Returns the unique virtual ID string for the positions coordinates"""
        m = hashlib.md5()
        for pos in positions:
            m.update(str(pos).encode())
        return m.hexdigest()

    @staticmethod
    def get_complete_virtual_id(actuator_logical_id: str, hashed_id: str) -> str:
        return actuator_logical_id + '_' + hashed_id

    @staticmethod
    def split_virtual_id(virtual_id: str) -> Tuple[str, str]:
        return tuple(virtual_id.split('_'))


@dataclasses.dataclass(frozen=True)
class PairedEdgesPayload:
    """An object to encapsulate the payload sent to the cloud
    when the system finished a devices pairing process."""
    devices: List[EdgeDescriptor]
    machines: List[MachineMetadata]
    tenant_id: str

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON.

        The Dictionary will contain the values required to re-create the instance from the JSON
        object in a pythonic way:
            object = MyObject( **json.loads(json_string) )
        """
        l = []
        machines = []
        for d in self.devices:
            l.append(d.reprJSON())

        for m in self.machines:
            machines.append(m.reprJSON())

        return dict(devices=l, machines=machines, tenant_id=self.tenant_id)

    @classmethod
    def build_from_dict(cls, d: Dict) -> Optional["PairedEdgesPayload"]:
        # All the known fields are in the dict
        if 'devices' not in d:
            return None

        edge_descriptors: List[EdgeDescriptor] = list(map(
            lambda edge_descriptor_d: EdgeDescriptor.build_from_dict(edge_descriptor_d),
            d['devices']
        ))

        machines: List[MachineMetadata] = list(map(
            lambda machine_dict: MachineMetadata.build_from_dict(machine_dict),
            d['machines']
        ))

        return cls(edge_descriptors, machines, d['tenant_id'])



"""
This module defines the Gateway class and other related class
which support the Gateway conceptualization and functionality.
"""
# ============================ imports =========================================
from txp.common.config import settings
from txp.common.json_complex_encoder import ComplexEncoder
from txp.devices.job_task_manager import JobTasksManager
from txp.devices.drivers.driver import StateReport, DriverState
from txp.devices.package import GatewayPackage
from txp.devices.package_queue import PackageQueueProxy, PackageQueue
from txp.devices.sampling_task import SamplingTask
from txp.devices.gateway_states_machine_mixin import GatewayStatesMachineMixin
from txp.common.edge import EdgeDescriptor, MachineMetadata, PairedEdgesPayload
from txp.devices.mqtt import GatewayMQTTClient, MQTTMessageType
from txp.devices.telemetry_sender import TelemetrySender
from txp.devices.ux.ux_client import UxClient
import txp.devices.ux.ux_states_protocol as ux_states
from txp.devices.devices_manager import DevicesManager
from collections import defaultdict
from deepdiff import DeepDiff
from txp.devices.production_manager import ProductionManager
from txp.common.configuration import JobConfig
from txp.common.configuration import GatewayConfig
from typing import List, Dict, Optional
from typing import List, Dict, Set, Optional, Tuple
from threading import Thread, Event
from queue import SimpleQueue
from multiprocessing import managers
from txp.devices.gateway_states_enums import *
from txp.common.local_ip import get_local_ip
from txp.devices.exceptions import *
from txp.devices.state_payload import GatewayStatePayload
import time
import json
import datetime
import logging
import requests

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Definition of the custom GatewayManager to handle server processes.
# ==============================================================================
class _GatewayManager(managers.BaseManager):
    pass


# ==============================================================================
# Definition of the custom Gateway class
# ==============================================================================
class Gateway(GatewayStatesMachineMixin):
    """This Gateway is the defined Gateway concept in the Tranxpert architecture.

    The Gateway takes care on instantiation of all the configured drivers, by interacting
    with the DriversHosts.

    The Gateway contains a PackageQueue which is shared to all the Driver instances.
    Each driver will drop packages on that queue, and the gateway defines its own
    strategy to consume those packages and send it to the cloud.

    The Gateway instance is designed to exists as soon as the main program process starts,
    and it contains an internal state machine which ensure that the system transitions
    from state to state, collecting all the information required by the gateway to operate.

    """

    def __init__(self):
        """
        Initializes the Gateway states machine.

        This init process annotates attributes of the Gateway object that are expected to be assigned
        when the state machine transitions through the execution flow.
        """
        super(Gateway, self).__init__()

        # Gateway package Queue.
        _GatewayManager.register("PackageQueue", PackageQueue, PackageQueueProxy)
        self._server_process_manager = _GatewayManager()
        self._server_process_manager.start()
        self.package_queue: PackageQueueProxy = (
            self._server_process_manager.PackageQueue()
        )

        self._ux_client: UxClient = UxClient()

        self._ux_client.start_streamlit_app()

        self._mqtt_client: GatewayMQTTClient = None
        """_mqtt_client: The internal MQTT client for the Gateway to use."""

        self._job_manager: JobTasksManager = None

        self._device_manager: DevicesManager = DevicesManager()
        """_device_manager: The DeviceManager is the communication mean between the Gateway and the 
        physical devices connected to it."""

        self._devices_state_table: Dict[str, StateReport] = {}
        """_devices_state_table: a dictionary with the most recent received state for an edge device
        in the DeviceManager"""

        self._production_manager: ProductionManager = None
        """_production_manager: The production manager takes charge of devices pairing in production phase"""

        self._machine_to_edges_table: Dict[str, List[str]] = {}
        """_machines_to_edges_table: a lookup table to get which edges ids are associated to a machine"""

        self._edge_to_machine_id_table: Dict[str, List[str]] = defaultdict(list)
        """_edge_to_machine_id_table: a table to lookup the asset/machine monitored by the edge with 
        logical ID. The value is a list because a virtual edge actuator can be associated with multiple edges."""

        self._machines_table_by_id: Dict[str, MachineMetadata] = {}

        self._edge_logical_id_to_edge_descriptor: Dict[str, EdgeDescriptor] = {}

        self._new_configuration_received_flag: bool = False
        """_new_configuration_received_flag: flag to track when a new configuration was received and some components 
        needs to update accordingly"""

        self._plain_received_configuration: Dict = {}
        """_plain_received_configuration: The Dictionary received by the cloud in the most recent configuration"""

        self._ping_received: bool = False
        """_ping_received: flag to indicate when a configuration was send without changes. This is interpreted
        as a PING request to the Gateway, who then will send the State to the cloud."""

        self._edges_state_diff: bool = False
        """_edges_state_diff: When this flag is True, the Gateway will publish a State report as consequence."""

        # MQTT attributes
        self._mqtt_gateway_commands_topic: str = ""
        self._mqtt_gateway_config_topic: str = ""
        self._mqtt_gateway_errors_topic: str = ""

        # Temporary job_task_finished queue to know when to transition from sampling to ready
        self._job_task_finished_queue: SimpleQueue = None

    def start(self):
        """This method start the Gateway states machine, and delegates the
        execution of the program to the Gateway.

        This method will compute which is the appropriate state to begin in,
        whether this being a new production phase for a new instance, or an
        existent instance that is already configured.

        """

        self._load_global_state()

        if self.global_state.state == GatewayStates.Startup:
            self.begin_startup()

        elif self.global_state.state == GatewayStates.Ready:
            log.info("The Gateway found a previous Global State. "
                     "Trying to pull a new cloud configuration before preparing devices.")
            self._mqtt_client: GatewayMQTTClient = GatewayMQTTClient(self, self._mqtt_on_message_callback)
            self._initialize_mqtt_attributes()

            start_wait = datetime.datetime.now()
            while not self._new_configuration_received_flag and not self._ping_received:
                self._check_connection()
                time.sleep(1)
                if (datetime.datetime.now() - start_wait).total_seconds() > 10:
                    log.warning(f"No configuration received from cloud. Using the previous configuration saved.")
                    break

            if self._new_configuration_received_flag:
                log.info(f"The Gateway received a new configuration update "
                         f"to version: {self.global_state.last_configuration_id}")
            else:
                log.info(f"Configuration has not changed. Using version: {self.global_state.last_configuration_id}")

            self._configuration_reception_completed()

        else:
            raise RuntimeError("Invalid state when calling Gateway start.")

    def stop(self, *args, **kwargs) -> None:
        """Stop method used as callback to the SIGTERM/SIGINT interruptions."""
        log.info("Gateway process was requested to stop. Executing graceful termination...")
        self.package_queue.close()
        exit(0)

    # =========================== Startup State Methods =============================
    def _step_into_startup(self) -> None:
        """Callback to call after begin_startup transition trigger

        This callback will execute itself until the introduced values are valid
        and the Gateway can move to the next state.

        """

        # The Gateway started a production phase. Create the marker file.
        self._create_initialized_marker_file()

        valid_startup_values: bool = False
        retry: bool = False
        deserialization_error: bool = False

        while not valid_startup_values:
            self.status = StartupStatus.RequestRequiredValues
            log.info(self._state_log_msg())

            # Sends new message to UX. If there was an error with previously
            # introduced values, then those Details are sent.
            self._send_startup_ux_message(retry, deserialization_error)

            ux_state_rec: ux_states.UxStartupState = self._ux_client.wait_for_ux_state()

            if not isinstance(ux_state_rec, ux_states.UxStartupState):
                raise RuntimeError(
                    f"Unexpected UxState received in Gateway: {ux_state_rec.state}"
                )

            try:
                valid_startup_values = GatewayMQTTClient.validate_connection_values(
                    ux_state_rec.project_id,
                    self.gateway_config.cloud_region,
                    ux_state_rec.registry_id,
                    ux_state_rec.gateway_id,
                )
            except ValueError as key_deserialization_error:
                log.error(f"ERROR: {key_deserialization_error}")
                deserialization_error = True
            if not valid_startup_values:
                retry = True

        self.gateway_config.gateway_id = ux_state_rec.gateway_id
        self.gateway_config.registry_id = ux_state_rec.registry_id
        self.gateway_config.project_id = ux_state_rec.project_id
        log.info("Gateway: Startup Files Validated")
        self._startup_completed()

    def _send_startup_ux_message(self, retry: bool, deserialization_error: bool):
        """Send a Startup message to the UI application.

        If it's a retry, that means there was some problem with
        the received values.
        """
        if deserialization_error:
            details = ux_states.UxDetails(
                ux_states.UxDetails.Kind.Error,
                settings.ux.startup.invalid_private_key_cert,
            )
        elif retry:
            details = ux_states.UxDetails(
                ux_states.UxDetails.Kind.Warning,
                settings.ux.startup.invalid_values_warning,
            )
        else:
            details = None

        ux_state_sent = ux_states.UxStartupState(
            self.state,
            self.status,
            details,
            self.gateway_config.gateway_id,
            self.gateway_config.registry_id,
            False,
            False,
            False,
            False,
            self.gateway_config.project_id
        )
        self._ux_client.send_state_to_ui(ux_state_sent)

    # ======================== Configuration Reception State Methods ======================
    def _step_into_configuration_reception(self):
        """Callback to be invoked after the _startup_completed_callback trigger.

        This callback will proceed to listen for cloud configurations for the
        Gateway and the devices bound to the Gateway. Specifically, in order:
            - Gateway configuration
            - Machines Metadata configuration
            - Edges configuration
        """
        self.status = ConfigurationReceptionStatus.EstablishCloudConnection
        log.info(self._state_log_msg())

        # Send current state to the UI process
        ux_state_sent = ux_states.UxConfigurationReceptionState(
            self.state,
            self.status,
            None,
            self.gateway_config,
            self.machines_configurations,
            self.edges_configurations,
            False,
        )
        self._ux_client.send_state_to_ui(ux_state_sent)

        self._mqtt_client: GatewayMQTTClient = GatewayMQTTClient(self, self._mqtt_on_message_callback)
        self._initialize_mqtt_attributes()

        self._mqtt_client.subscribe_to_topic(
            self.gateway_config.gateway_id, self._mqtt_gateway_config_topic, False
        )

        self._mqtt_client.subscribe_to_topic(
            self.gateway_config.gateway_id, self._mqtt_gateway_errors_topic, False
        )

        start_wait = datetime.datetime.now()
        while not self._new_configuration_received_flag:
            log.info(f"Waiting for configuration to download in state {self.global_state.state}")

            try:
                self._mqtt_client.enable_incoming_messages_processing(1)
                time.sleep(1)
            except (MqttNetworkingUnavailableError, MqttConnectTimeoutError):
                log.error("Error detected in MQTT Connection while waiting configuration reception.")

            if (datetime.datetime.now() - start_wait).total_seconds() > 10:
                log.error(f"Timeout exceeded while waiting from configuration in state {self.global_state.state}")
                self._production_phase_not_completed()

        while not self.global_state.last_configuration_id != "0":
            log.info(f"Waiting for configuration to process in state {self.global_state.state}")
            time.sleep(1)
            if (datetime.datetime.now() - start_wait).total_seconds() > 10:
                log.error(f"Timeout exceeded while processing received configuration in state {self.global_state.state}")
                self._production_phase_not_completed()

        log.info(f"Configuration successfully processed in state {self.global_state.state}")

        self._configuration_reception_completed()

    # ======================== Configuration Validation State Methods =====================
    def _step_into_configuration_validation(self):
        self.status = ConfigurationValidationStatus.CheckConfigurations
        self._new_configuration_received_flag = False
        ux_state_sent = ux_states.UxConfigurationValidationState(
            self.state,
            self.status,
            None,
            self.gateway_config,
            self.machines_configurations,
            self.edges_configurations,
        )

        # Initialize different in-memory data
        self._build_internal_tables()

        self._ux_client.send_state_to_ui(ux_state_sent)

        num_paired_devices = self._get_number_of_paired_devices()
        log.info(f"Number of paired devices received in configuration: {num_paired_devices}")

        if num_paired_devices != 0 and num_paired_devices != len(self.edges_configurations):
            log.warning("Received non-zero and non-complete set of paired devices.")
            log.error("Partial devices pairing is not supported yet. Moving to error.")
            self._production_phase_not_completed()

        if num_paired_devices == 0:
            log.warning("Received 0 paired devices. Device pairing required. Gateway will move to device pairing")
            self._configuration_validation_completed()  # trigger to device pairing state

        if num_paired_devices == len(self.edges_configurations):
            log.info("All received devices in configuration are paired. Moving to Ready.")
            self._device_pairing_completed()  # Move to device preparing

    # ======================== Devices Pairing State Methods ==============================
    def _step_into_devices_pairing(self):
        """Callback to call when the UX is pairing devices."""

        # We need to check mqtt connection in order to keep alive
        stop_check: Event = Event()
        check_mqtt_conn_thread: Thread = Thread(
            name="check_mqtt_connect_prod",
            daemon=True,
            target=self._check_mqtt_conn_prod,
            args=(stop_check,)
        )
        check_mqtt_conn_thread.start()

        self._save_global_state()
        self._create_production_manager()
        self.status = DevicePairingStatus.SelectDeviceToPair
        ux_state_sent = self._get_ux_device_pairing_initial_state()
        self._ux_client.send_state_to_ui(ux_state_sent)

        # devices pairing information exchange loop with user interface interactions
        while True:
            ux_state = self._ux_client.wait_for_ux_state()

            if isinstance(ux_state, ux_states.UxDevicePairingState):
                device_to_pair_id = ux_state.props['device_to_pair']
                edge_descriptor_for_edge = self._edge_logical_id_to_edge_descriptor[
                    device_to_pair_id
                ]

                if not edge_descriptor_for_edge.is_actuator_device():
                    self._pair_non_virtual_device(
                        ux_state, edge_descriptor_for_edge
                    )
                else:
                    self._pair_virtual_device(
                        ux_state, edge_descriptor_for_edge
                    )

                if self._get_number_of_paired_devices() == len(self.edges_configurations):
                    log.info("All devices were successfully paired."
                             "Leaving the Devices Pairing state")
                    break

            else:
                log.error("Unexpected Error in Gateway. Unknown UX State received.")
                raise RuntimeError("Unexpected Error in Gateway. Unknown UX State received.")

        stop_check.set()
        self._production_manager = None
        self._publish_paired_devices_to_cloud()

        # Wait until the cloud updated the configuration after pairing
        while not self._ping_received:
            log.info("Waiting for cloud configuration acknowledge of paired devices")
            try:
                self._mqtt_client.enable_incoming_messages_processing(2)
            except (MqttNetworkingUnavailableError, MqttConnectTimeoutError):
                log.error("Error detected with MQTT client connection while waiting for "
                          "devices pairing acknowledge from cloud.")
                # TODO: This should put the Gateway in a Irrecoverable state in the States Machine.

        log.info("Received cloud acknowledge of new paired configuration")

        self._device_pairing_completed()

    def _check_mqtt_conn_prod(self, stop: Event):
        while not stop.is_set():
            self._check_connection()

        log.info("Closed check mqtt connection in production")

    def _get_ux_device_pairing_initial_state(self):
        """Returns the initial ux_state for the DevicesPairing state."""
        edges_to_pair = []
        for machine_metadata in self.machines_configurations:
            edges_to_pair.extend(machine_metadata.edges_ids)
        expected_devices = set(edges_to_pair)
        props = {
            "paired_devices": set(),
            "expected_devices": expected_devices,
            "device_to_pair": None,
            "virtual_edge_asset": None
        }
        return ux_states.UxDevicePairingState(
            self.state, self.status, None, self.gateway_config, self.machines_configurations,
            self.edges_configurations, props
        )

    def _pair_non_virtual_device(self, ux_state, edge_descriptor: EdgeDescriptor):
        """This method interacts with the ProductionManager when the device to pair is not virtual"""
        log.info(f"Pairing Device with logical ID: {edge_descriptor.logical_id}")
        ux_state_sent = ux_state.copy()

        ux_state_sent.status = DevicePairingStatus.PairingSelectedDevice  # Update UX to inform the user
        self._ux_client.send_state_to_ui(ux_state_sent)

        self._production_manager.start_listening_for_an_edge(edge_descriptor.device_kind)
        device_wait_result = self._production_manager.wait_for_device_to_connect(
            edge_descriptor.device_kind, 3
        )  # TODO it says minutes. It's really minutes or seconds?

        self._production_manager.stop_listening_for_devices(edge_descriptor.device_kind)

        if device_wait_result:
            device = self._production_manager.get_hosted_device()
            log.info(f"Production Manager obtained device: {device}")

            ux_state_sent.status = DevicePairingStatus.RequestConfirm  # Update UX to request confirmation from user
            self._ux_client.send_state_to_ui(ux_state_sent)

            ux_state = self._ux_client.wait_for_ux_state()  # Wait for user confirmation

            if ux_state.status == DevicePairingStatus.PairingConfirmed:
                log.info("Pairing was confirmed")
                ux_state_sent = ux_state.copy()
                ux_state_sent.status = DevicePairingStatus.SelectDeviceToPair
                ux_state_sent.props['paired_devices'] = (
                    ux_state_sent.props['paired_devices'].union({edge_descriptor.logical_id})
                )
                self._ux_client.send_state_to_ui(ux_state_sent)

                # Update internal configuration with discovered physical ID
                edge_descriptor.set_physical_id(device[1].physical_id)

                log.info(
                    f"Physical ID {device[1].physical_id} added in local configuration "
                    f"for edge: {edge_descriptor.logical_id}"
                )

            elif ux_state.status == DevicePairingStatus.PairingCancelled:
                log.info("Pairing was cancelled")
                ux_state_sent = ux_state.copy()
                ux_state_sent.status = DevicePairingStatus.SelectDeviceToPair  # If cancelled, restart state to select
                self._ux_client.send_state_to_ui(ux_state_sent)

            else:
                log.error("Unexpected Error in Gateway. Unknown UX State received.")
                raise RuntimeError("Unexpected Error in Gateway. Unknown UX State received.")

        else:
            log.error("Error while waiting for device pairing")
            ux_state_sent = ux_state.copy()
            ux_state_sent.status = DevicePairingStatus.PairingError
            self._ux_client.send_state_to_ui(ux_state_sent)

    def _pair_virtual_device(self, ux_state, edge_descriptor: EdgeDescriptor):
        log.info(f"Pairing Virtual Edge arm with logical ID: {edge_descriptor.logical_id}")
        ux_state_sent = ux_state.copy()

        self._production_manager.prepare_virtual_production(edge_descriptor)

        while True:
            next_status_step: DevicePairingStatus = self._production_manager.get_next_virtual_production_step()
            ux_state_sent.props['virtual_production'] = self._production_manager.get_virtual_production_record()
            ux_state_sent.status = next_status_step
            log.info(f"Sent status: {next_status_step} to virtual production UI")
            self._ux_client.send_state_to_ui(ux_state_sent)  # Update UX to request confirmation from user

            ux_state = self._ux_client.wait_for_ux_state()  # Wait for user interaction

            if ux_state.status == DevicePairingStatus.VirtualCameraConnectionConfirmed:
                log.info("User confirmed virtual camera connection. Trying to detect device")
                camera_connected_status = self._production_manager.detect_virtual_camera()
                ux_state_sent = ux_state.copy()
                ux_state_sent.status = camera_connected_status
                self._ux_client.send_state_to_ui(ux_state_sent)
                time.sleep(5)
                continue

            if ux_state.status == DevicePairingStatus.VirtualThermocameraConnectionConfirmed:
                log.info("User confirmed virtual thermo camera connection. Trying to detect device")
                thermocamera_connected_status = self._production_manager.detect_virtual_thermocamera()
                ux_state_sent = ux_state.copy()
                ux_state_sent.status = thermocamera_connected_status
                self._ux_client.send_state_to_ui(ux_state_sent)
                time.sleep(15)  # Sleeps for the user to connect cameras after production
                continue

            else:  # at this point the user will interact with the arm robot in a positions capture loop
                log.info(f"Received requested user next step in virtual production: {ux_state.status}")
                if ux_state.status == DevicePairingStatus.VirtualEnableTorque:
                    self._production_manager.enable_torque()

                if ux_state.status == DevicePairingStatus.VirtualDisableTorque:
                    self._production_manager.disable_torque()

                if ux_state.status == DevicePairingStatus.VirtualCapturePosition:
                    self._production_manager.capture_virtual_pos()

                if ux_state.status == DevicePairingStatus.VirtualRejectCapturedPosition:
                    self._production_manager.reject_captured_pos()

                if ux_state.status == DevicePairingStatus.VirtualConfirmCapturedPosition:
                    virtual_edge_asset = ux_state.props["virtual_edge_asset"]
                    self._production_manager.confirm_captured_pos(edge_descriptor, virtual_edge_asset)

                if ux_state.status == DevicePairingStatus.VirtualConfirmAllPositions:
                    virtual_pos_tuple = self._production_manager.confirm_all_captured_pos(edge_descriptor)
                    virtual_descriptors = [entry[0] for entry in virtual_pos_tuple]
                    machines = [entry[1] for entry in virtual_pos_tuple]

                    self.edges_configurations = self.edges_configurations + virtual_descriptors

                    for i, machine_id in enumerate(machines):
                        self._machines_table_by_id[machine_id].edges_ids = (
                            self._machines_table_by_id[machine_id].edges_ids +
                            [ virtual_descriptors[i].logical_id ]
                        )

                    ux_state_sent = ux_state.copy()
                    ux_state_sent.status = DevicePairingStatus.SelectDeviceToPair
                    ux_state_sent.props['paired_devices'] = (
                        ux_state_sent.props['paired_devices'].union({edge_descriptor.logical_id})
                    )
                    self._ux_client.send_state_to_ui(ux_state_sent)
                    break
        self._build_internal_tables()  # Refresh internal tables after virtual pairing
        self._production_manager.finish_virtual_production()

    def _publish_paired_devices_to_cloud(self) -> None:
        """This method will publish the paired EdgeDescriptor's to the cloud.

        The cloud will take charge of finishing the pairing in the project model.
        """
        try:
            payload = PairedEdgesPayload(
                self.edges_configurations, self.machines_configurations, self.gateway_config.tenant_id
            )  # Not deep copy required
            payload_json = json.dumps(payload, cls=ComplexEncoder, indent=2)
            gateway_devices_pairing_topic = (
                settings.gateway.mqtt.devices_pairing_topic_template.format(
                    self.gateway_config.gateway_id
                )
            )

            self._mqtt_client.publish_message(
                MQTTMessageType.DEVICES_PAIRING_MESSAGE,
                self.gateway_config.gateway_id,
                gateway_devices_pairing_topic,
                payload_json,
                False,
                False
            )

            log.info(f"Publishing paired devices information in MQTT Topic: {gateway_devices_pairing_topic}")
            log.info(f"Published PairedEdgesPayload: {payload_json}")
        except Exception as e:
            log.error(f'Unexpected error while trying to publish paired devices information event. {e}')

    # ======================== Preparing Devices State Methods ==============================
    def _step_into_preparing_devices(self):
        """Callback to be called when the Gateway enters in the Preparing Devices State.

        It will create the DriversHosts, the Drivers, and wait for the devices to connect.

        TODO: Handle possible exceptions in performed tasks by this method.
        TODO: Transition to specific error when some operation failed. That error should
            be communicated to the UI.
        """
        self.global_state.collected_packages_by_edge = dict(map(
            lambda edge_descriptor: (edge_descriptor.logical_id, 0),
            self.edges_configurations
        ))

        # Clear drivers hosts and wait before creating new hosts
        # to avoid state report too fast.
        self._device_manager.clear_devices()
        time.sleep(1.5)

        self.status = PreparingDevicesStatus.CreatingDriversHosts

        self._update_ux_preparing_state(settings.ux.preparing_devices.configuring_devices_info)

        # Create the DriversHost instances
        self._create_drivers_hosts()

        # Create the Drivers Instances
        self.status = PreparingDevicesStatus.CreatingDrivers
        self._create_drivers()

        self._initialize_job_manager()
        self._job_manager.start()

        self.status = PreparingDevicesStatus.WaitingForDevicesToBeConnected

        self._wait_for_drivers_to_be_ready()

        self._preparing_devices_completed()

    def _update_ux_preparing_state(self, message: str) -> None:

        details = ux_states.UxDetails(
            ux_states.UxDetails.Kind.Info,
            message
        )
        ux_state_sent = ux_states.UxPreparingDevicesState(
            self.state, self.status, details,
            self.gateway_config, self.machines_configurations, self.edges_configurations,
            self._devices_state_table
        )

        self._ux_client.send_state_to_ui(ux_state_sent)

    # ======================== Ready - Sampling States Methods ==================================
    def _step_into_ready_state(self):
        """Callback to be called when the Gateway enters the Ready State"""
        self.status = ReadyStatus.Idle
        log.info('Gateway entered the Ready state')
        self._create_ready_marker_file()
        self._save_global_state()

        self._update_edges_states()
        self._publish_states()
        time.sleep(1)
        self._ping_received = False

        while True:
            self._update_ux_ready_state(settings.ux.ready_state.gateway_is_ready)

            if not settings.gateway.offline_mode:
                self._check_connection()

            if self._at_least_one_edge_sampling():
                log.info('Sampling Window in progress. Gateway transition to Sampling')
                self._start_sampling()

            if self._new_configuration_received_flag:
                log.info('Gateway new configuration received. '
                         'The Gateway will proceed to update to the new configuration.')
                self._job_manager.stop()

                log.debug("Gateway moving to Preparing Devices after new configuration received")
                self._configuration_reception_completed()
                # transitions to configuration reception completed

            elif self._ping_received:
                log.debug("Configuration PING received. Sending states to cloud.")
                self._publish_states()
                self._ping_received = False

            elif self._edges_state_diff:
                log.debug("Flag _edges_state_diff holds True. Publishing State Report.")
                self._publish_states()
                self._edges_state_diff = False

            self._update_edges_states()
            time.sleep(5)

    def _update_ux_ready_state(self, message: str) -> None:
        details = ux_states.UxDetails(
            ux_states.UxDetails.Kind.Info,
            message
        )
        ux_state_sent = ux_states.UxReadyState(
            self.state, self.status, details, self.gateway_config,
            self.machines_configurations, self.edges_configurations,
            self.global_state.packages_collected, self.global_state.packages_sent,
            self.global_state.packages_collected, self.global_state.collected_packages_by_edge,
            self._devices_state_table
        )
        self._ux_client.send_state_to_ui(ux_state_sent)

    def _step_into_sampling_state(self):
        """
           The Gateway should keep sampling until some condition happen.
               - Someone commanded the Gateway to Stop.
               - The Gateway detected some condition inside the system that
                   requires the sampling to stop.
        """

        self.status = GatewaySamplingStatus.Sampling
        close_queue_interval = datetime.datetime.now()

        self._update_edges_states()
        self._publish_states()
        time.sleep(1)  # Waits for 1 sec to avoid state publish limit of 1/sec

        while True:
            # TODO: Commenting this break condition because it can overflow the transitions library
            # if self._check_if_edges_finished_sampling():
            #     log.info("Edges finished current task")
            #     self._device_manager.stop_sampling_windows()
            #     break

            if (
                self._job_task_finished_queue.qsize()
                and self._check_if_edges_finished_sampling()
            ):
                task_window_manager_id = self._job_task_finished_queue.get()
                log.info(f"Job Task reported as finished {task_window_manager_id}. "
                         f"Leaving sampling state")

                while self._job_task_finished_queue.qsize():
                    self._job_task_finished_queue.get()

                self._sampling_completed()

            if self._new_configuration_received_flag:
                log.info("New configuration received. Requesting to stop the current sampling")
                self._device_manager.stop_sampling_windows()

                while not self._check_if_edges_finished_sampling():
                    time.sleep(1)

                log.info('The sampling windows has been stopped.')
                break

            log.debug("Waiting for edges to finish the current task.".format(self._devices_state_table))
            if not settings.gateway.offline_mode:
                try:
                    self._check_connection()
                except (MqttNetworkingUnavailableError, MqttConnectTimeoutError):
                    log.error("Error detected with MQTT client while checking connection")

            self._update_edges_states()

            if self._edges_state_diff:
                log.info("Flag _edges_state_diff holds True. Publishing State Report.")
                self._publish_states()
                self._edges_state_diff = False

            elif self._ping_received:
                log.info("Configuration PING received. Sending states to cloud.")
                self._publish_states()
                self._ping_received = False

            # secure current queue disk checkpoint
            elapsed = datetime.datetime.now()
            if ((elapsed - close_queue_interval).total_seconds() >
                    settings.gateway.queue_max_time_to_persist):
                close_queue_interval = datetime.datetime.now()
                self.package_queue.save_current_disk_state()

            time.sleep(5)

        self.status = GatewaySamplingStatus.CollectPackages
        self._sampling_completed()

    def _check_connection(self):
        """Checks the health of the MQTT connection.

        This will:
            - Check the health of the connection.
            - Process any callback for incoming messages that are pending.
        """
        # Call to subscribe to topics, just in case if there was a networking error at initialization
        try:
            self._mqtt_client.subscribe_to_topic(
                self.gateway_config.gateway_id, self._mqtt_gateway_config_topic, False
            )

            self._mqtt_client.subscribe_to_topic(
                self.gateway_config.gateway_id, self._mqtt_gateway_errors_topic, False
            )
            self._mqtt_client.enable_incoming_messages_processing(1)
        except (MqttNetworkingUnavailableError, MqttConnectTimeoutError):
            log.error("Error detected with MQTT client while checking connection")

    # ======================= Error transitions callbacks ===========================
    def _step_into_production_not_completed_error(self):
        """Callback to call when the _production_phase_not_completed trigger is invoked.

        If this callback is called, it meas that the Gateway never completed the Production
        phase, and it will be unusable until restarted outside of the production line.
        """
        self.status =GatewayErrorStatus.ProductionNotCompleted
        details = ux_states.UxDetails(
            ux_states.UxDetails.Kind.Error,
            settings.ux.error.production_not_completed
        )
        ux_state_sent = ux_states.UxErrorState(
            self.state,
            self.status,
            details,
        )

        self._ux_client.send_state_to_ui(ux_state_sent)

        self._ux_client.wait_for_ux_state()  # Hard Block. The Gateway will not operate anymore.

    # ======================= Private helper methods =======================================
    def _parse_gateway_configuration_object(self, config: Dict) -> None:
        """Parses the Job configuration for the Gateway received from cloud.

        This object contains the following first-level attributes:
            - "job": the job configuration for the Gateway operation.
            - "gateway": The GatewayConfig instance for the Gateway
            - "edges": The edges configuration
            - "machines": The machines configurations.
        """
        root_required_fields = [
            "job", "gateway", "machines", "edges", "configuration_id"
        ]

        for field in root_required_fields:
            if field not in config:
                log.error(f'Configuration does not have {field} field. Could not process received configuration')
                return

        if self.global_state.last_configuration_id != config['configuration_id']:  # Configuration Changed

            # Checks if the configuration changed values. If not, then updates only the configuration version
            if not self._check_configuration_diff(config):
                log.info("A new configuration version received, but the values didn't change for this Gateway."
                         " Updating version number")
                self.global_state.last_configuration_id = config['configuration_id']
                log.info(f"New version number set to: {self.global_state.last_configuration_id}")
                return

            # Validates the Job object
            job_config = JobConfig.build_from_dict(config['job'])
            if not job_config:
                log.error("The Gateway could not create the new Job configuration instance. "
                          "Configuration not updated")
                return

            # Validates the GatewayConfig object
            gateway_config = GatewayConfig.build_from_dict(config['gateway'])
            if not gateway_config:
                log.error("The Gateway could not create the new GatewayConfig instance. "
                          "Configuration not updated")
                return

            # Validates the edges
            edges: List[EdgeDescriptor] = []
            for edge_obj in config['edges']:
                edge_descriptor = EdgeDescriptor.build_from_dict(edge_obj)
                if not edge_descriptor:
                    log.error("The Gateway could not create the associated EdgeDescriptor for "
                              f"the edge: {edge_obj.get('logical_id', 'No logical_id found.')}")
                edges.append(edge_descriptor)

            if any(
                list(map(
                    lambda edge: edge is None, edges
                ))
            ):
                return

            # Validates the machines
            machines: List[MachineMetadata] = []
            for machine_obj in config['machines']:
                machine_metadata = MachineMetadata.build_from_dict(machine_obj)
                if not machine_metadata:
                    log.error("The Gateway could not create the associated MachineMetadata for "
                              f"the machine: {machine_obj.get('machine_id', 'No machine_id found.')}")
                machines.append(machine_metadata)

            if any(
                list(map(
                    lambda machine: machine is None, machines
                ))
            ):
                return

            # Note: Additional validations can be performed here.

            log.info(f'New Job received from cloud. Configuration ID: {config["configuration_id"]}')
            log.info(f'New Gateway Configuration received from cloud with machines: {gateway_config.machines_ids}')
            log.info(f'Received {len(edges)} edges in the configuration')
            log.info(f'Received {len(machines)} machines in the configuration')

            self.global_state.last_configuration_id = config["configuration_id"]
            self.gateway_config.machines_ids = gateway_config.machines_ids
            self.gateway_config.job = job_config
            self.machines_configurations = machines
            self.edges_configurations = edges
            self.gateway_config.tenant_id = gateway_config.tenant_id
            self._plain_received_configuration = config
            self._new_configuration_received_flag = True

        else:
            log.info('Configuration received but version has not changed.')
            if not self._new_configuration_received_flag:
                self._ping_received = True

            if not self._plain_received_configuration:
                self._plain_received_configuration = config

    def _check_configuration_diff(self, new_conf: Dict) -> bool:
        """Helper method to check if a received configuration has changes and requires updating
        the current active configuration.

        Returns:
            True if the configuration changed.
        """
        received_config = new_conf.copy()
        received_config.pop("configuration_id")
        previous_received_conf = self._plain_received_configuration.copy()
        previous_received_conf.pop("configuration_id", None)
        return bool(DeepDiff(received_config, previous_received_conf))

    def _build_internal_tables(self) -> None:
        """Build some internal in-memory tables used by the Gateway when a new configuration
        starts to operate."""
        # Initializes States dictionary for the configured edges
        self._devices_state_table.clear()
        for edge_descriptor in self.edges_configurations:
            if edge_descriptor.is_actuator_device():
                continue

            self._devices_state_table[edge_descriptor.logical_id] = StateReport(
                edge_descriptor.logical_id, DriverState.DISCONNECTED
            )

        # Initializes machines to edges dictionary based on configuration
        self._machine_to_edges_table.clear()
        self._machines_table_by_id.clear()
        self._edge_to_machine_id_table.clear()
        self._edge_logical_id_to_edge_descriptor.clear()
        for machine_metadata in self.machines_configurations:
            self._machine_to_edges_table[machine_metadata.machine_id] = machine_metadata.edges_ids
            self._machines_table_by_id[machine_metadata.machine_id] = machine_metadata

            for edge_serial in machine_metadata.edges_ids:
                self._edge_to_machine_id_table[edge_serial].append(machine_metadata.machine_id)

        for edge_des in self.edges_configurations:
            self._edge_logical_id_to_edge_descriptor[edge_des.logical_id] = edge_des

    def _get_number_of_paired_devices(self) -> int:
        """Returns the number of paired devices in the configuration.

        The criteria of pairing is provided by the DeviceId dataclass.
        """
        paired_devices = list(map(lambda edge_des: edge_des.is_paired(), self.edges_configurations))
        num_paired_devices = paired_devices.count(True)
        return num_paired_devices

    def _initialize_job_manager(self) -> None:
        telemetry_sender = TelemetrySender(
            self.package_queue, self._mqtt_client, self._process_package_stats,
            None, self.gateway_config,
            list(map(
                lambda descriptor: descriptor.logical_id, self.edges_configurations
            ))
        )

        self._job_task_finished_queue = SimpleQueue()

        self._job_manager = JobTasksManager(
            self.gateway_config,
            telemetry_sender,
            self._device_manager,
            self.global_state.last_configuration_id,
            self._machines_table_by_id,
            self._job_task_finished_queue
        )

    def _create_production_manager(self) -> ProductionManager:
        """Returns the instance of the Production Manager to use in device pairing"""
        required_drivers_host = set(map(
            lambda edge_descriptor: edge_descriptor.device_kind, self.edges_configurations
        ))
        log.info(f"Production Manager will configure the hosts: {required_drivers_host}")
        self._production_manager = ProductionManager(self.package_queue, list(required_drivers_host))

    def _create_drivers_hosts(self) -> None:
        """Request to the DeviceManager the creation of the required DriversHost"""
        required_drivers_host = set(map(
            lambda edge_descriptor: edge_descriptor.device_kind, self.edges_configurations
        ))
        log.info(f"DriversHost to configure: {required_drivers_host}")

        for drivers_host_type in required_drivers_host:
            self._device_manager.create_host(drivers_host_type, self.package_queue)
            log.info(f"DriversHost created: {drivers_host_type}")

    def _create_drivers(self) -> None:
        """Request to the DeviceManager to create the appropriate Drivers instances
        for the configured edge devices"""
        # Creation of the List of Tuple[EdgeDescriptor, SamplingWindow] to create the Drivers
        self._device_manager.initialize_drivers(self.edges_configurations)
        self._device_manager.start()

        log.debug("Drivers Added to DriversHost")

    def _update_edges_states(self) -> None:
        """Update the edges states table by reading the states reported by the driver.

        If at least one edge changed its previous stored state, then the flag
        _edges_state_diff will change to True.
        """
        received_states = self._device_manager.get_state_changed_queued_items()
        for state_report in received_states:
            if state_report.driver_state != self._devices_state_table[state_report.logical_id]:
                self._devices_state_table[state_report.logical_id] = state_report
                self._edges_state_diff = True


    def _wait_for_drivers_to_be_ready(
            self, timeout_scs: int = settings.gateway.wait_for_devices_startup_timeout
    ) -> None:
        """Waits for all edges to be ready to sample.
        """

        log.info("Waiting for all the devices to be CONNECTED.")
        started = datetime.datetime.now()
        while not all(
                list(map(
                    lambda state_report: state_report.driver_state == DriverState.CONNECTED,
                    self._devices_state_table.values()
                ))
        ):
            self._update_edges_states()
            self._update_ux_preparing_state(settings.ux.preparing_devices.waiting_for_devices_to_connect)

            current_time = datetime.datetime.now()
            if (current_time - started).total_seconds() > timeout_scs:
                log.info('Timeout for devices to connect exceeded. Starting Gateway anyways.')
                return

            else:
                time.sleep(5)

        log.info("All expected edge devices are CONNECTED")

    def _check_if_edges_finished_sampling(self) -> bool:
        """Returns True if all edges drivers are in any state different than SAMPLING.
        The queue with the driver state will report this change.
        """
        self._update_edges_states()
        return all(
            list(map(
                lambda state_report: state_report.driver_state != DriverState.SAMPLING,
                self._devices_state_table.values()
            )))

    def _at_least_one_edge_sampling(self) -> bool:
        """Returns True if at least an edge is sampling, which means that some Sampling
        Window has started"""
        value = any(list(map(
                    lambda state_report: state_report.driver_state == DriverState.SAMPLING,
                    self._devices_state_table.values()
                )))
        return value

    def _mqtt_on_message_callback(self, unused_client, unused_userdata, message):
        """Callback to be executed when a message is received in some subscribed topic
        in the MQTT client"""
        log.info(f"Gateway Message received in MQTT in topic: {message.topic}")

        if message.topic == self._mqtt_gateway_config_topic:
            try:
                config = json.loads(message.payload.decode("utf-8"))
            except json.decoder.JSONDecodeError as e:
                log.error('Error parsing received configuration JSON. Configuration update ignored')
                return

            log.info('Gateway configuration received from Cloud')
            self._parse_gateway_configuration_object(config)

        elif message.topic == self._mqtt_gateway_errors_topic:
            try:
                error = json.loads(message.payload.decode("utf-8"))
            except json.decoder.JSONDecodeError as e:
                log.error('Error parsing received error message JSON.')
                return

            log.error(f'Error received from cloud. {error["error_type"]}, '
                      f'caused by device {error["device_id"]}')

            if error["error_type"] == 'GATEWAY_ATTACHMENT_ERROR':
                log.warning('GATEWAY_ATTACHMENT_ERROR can happens because the device does not exist in IoT Core.')
        else:
            log.warning("Unknown Message received from Cloud.")

    def _scheduled_task_callback(self, machines_tasks: Dict[str, Dict[str, SamplingTask]] ):
        """The callback to be called for each scheduled tasks. This callback is forwarded to the TasksScheduler
        hold by the Gateway."""
        log.info(f"Scheduled sampling tasks executed at: {datetime.datetime.now()}")
        for machine, sampling_tasks in machines_tasks.items():
            self._device_manager.receive_sampling_tasks(sampling_tasks)
            self._device_manager.start_sampling_windows()

    def _process_package_stats(self, package: GatewayPackage, sent_to_cloud: bool) -> None:
        """Process the collected package and register the appropriate stats in the
        Global State"""
        self.global_state.packages_collected = self.global_state.packages_collected + 1
        self.global_state.collected_packages_by_edge[package.metadata.edge_descriptor.logical_id] += 1

        if sent_to_cloud:
            self.global_state.packages_sent = self.global_state.packages_sent + 1

    def _initialize_mqtt_attributes(self) -> None:
        """This method will initialize the MQTT related attributes hold by
        the gateway to be used in the callback when receiving a message.

        TODO: This attributes should be initialized in the production phase
            and should be stored in the global state.
            Currently this attributes are initialized after reading the hardwired
                configuration modules.
        """
        self._mqtt_gateway_commands_topic = settings.gateway.mqtt.device_commands_topic_template.format(
            self.gateway_config.gateway_id
        )
        self._mqtt_gateway_config_topic = settings.gateway.mqtt.device_config_topic_template.format(
            self.gateway_config.gateway_id
        )
        self._mqtt_gateway_errors_topic = settings.gateway.mqtt.device_error_topic_template.format(
            self.gateway_config.gateway_id
        )

    def _get_ngrok_public_address(self) -> Optional[str]:
        address = ""
        try:
            r = requests.get(settings.gateway.ngrok_local_address)
            address = json.loads(r.content.decode())['tunnels'][0]['public_url']
        except Exception as e:
            log.error(f"Error when requesting NGROK local server: {e}")
            address = None
        finally:
            return address

    def _get_state_payload(self) -> GatewayStatePayload:
        """Returns the Dictionary with the desired state fields to report to cloud."""
        payload = GatewayStatePayload(
            self.state,
            list(self._devices_state_table.values()),
            self.global_state.boot_time.strftime(
                settings.gateway.tasks_scheduler.start_date_format
            ),
            self._get_ngrok_public_address(),
            get_local_ip(),
            self.gateway_config.tenant_id,
        )
        return payload

    def _publish_states(self) -> None:
        """This method will publish the state of the gateway and the states of the edge
        devices.
        """
        try:
            state_json_obj = json.dumps(
                self._get_state_payload(), indent=4, cls=ComplexEncoder
            )

            gateway_state_topic = settings.gateway.mqtt.device_state_topic_template.format(
                self.gateway_config.gateway_id
            )
            self._mqtt_client.publish_message(
                MQTTMessageType.STATE_MESSAGE,
                self.gateway_config.gateway_id,
                gateway_state_topic,
                state_json_obj,
                False,
                False
            )

            for edge_serial, state_report in self._devices_state_table.items():
                edge_state = state_report.driver_state
                log.debug(f'Publishing State event for Driver: {edge_serial}')
                edge_topic = settings.gateway.mqtt.device_state_topic_template.format(
                    edge_serial
                )
                self._mqtt_client.publish_message(
                    MQTTMessageType.STATE_MESSAGE,
                    edge_serial,
                    edge_topic,
                    json.dumps({"state": str(edge_state)}),
                    attach=True, track_acknowledge=False
                )
        except Exception as e:
            log.error(f'Unexpected error while trying to publish the Gateway state: {e}')

    # ======================= Legacy Methods ========================================
    def _state_log_msg(self):
        return f"Gateway enters to {self.state} with status {self.status}"



import numpy as np
from common import iCOMOX_messages
from scipy.signal import welch
from scipy.signal.windows import hamming
import struct

class class_ADXL362:
    def __init__(self):
        # Initializing Fs_Acc, MFFT_Acc, window
        TicksInSamplingTimePeriod = int(0.5 + (0.0025 / (1/32768)))
        self.Fs_Acc = 32768 / TicksInSamplingTimePeriod    # 400 Hz
        self.NFFT_Acc = 512 #iCOMOX_messages.ADXL362_SAMPLES_NUM_PER_PIN    # 512
        self.Sensitivity_Acc = 1/1024  # LSB = 1/1024 [g] for the +/-2g range
        self.window = hamming(M=self.NFFT_Acc)

    def ADC_readings_to_g_units(self, ADC_samples):
        # convert ADC samples to g units
        return np.array(ADC_samples) * self.Sensitivity_Acc

    def g_units_to_PSD(self, acc_units_g):
        (freq_Acc, Pxxf_Acc) = welch(x=acc_units_g, fs=self.Fs_Acc, window=self.window, noverlap=self.NFFT_Acc/2, nfft=self.NFFT_Acc, return_onesided=True, detrend=False)
        return freq_Acc, Pxxf_Acc

    def msg_bytes_to_samples_g(self, payload):
        ADC_samples = np.array(struct.unpack("<{}h".format(iCOMOX_messages.ADXL362_SAMPLES_NUM*3), payload))
        ADC_samples = self.ADC_readings_to_g_units(ADC_samples=ADC_samples)
        return ADC_samples[0::3], ADC_samples[1::3], ADC_samples[2::3]
        # axis_len = len(ADC_samples) // 3
        # return ADC_samples[0:axis_len], ADC_samples[axis_len:2*axis_len], ADC_samples[2*axis_len:]


"""
This module implements a generic driver host for devices attached physically to the gateway.
"""
# ============================ imports =========================================
import multiprocessing
from typing import List, Tuple, Dict
from txp.devices.drivers.driver import DriverState, StateReport
from txp.common.edge import EdgeDescriptor
from txp.devices.drivers.drivers_host import DriversHost
from txp.devices.drivers.thermocouple.mcp9600_driver import Mcp9600Driver
from txp.devices.package_queue import PackageQueueProxy
from txp.common.edge.device_id import DeviceId
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# _GeneralDriversRegistry internal dictionary to map usb drivers string name to the
# concrete implementation.
# ==============================================================================
_GeneralDriversRegistry: Dict = {
    Mcp9600Driver.__name__: Mcp9600Driver
}

# ==============================================================================
# GenericDriversHost
# ==============================================================================


class GenericDriversHost(DriversHost):
    """Concrete implementation of a GenericDriversHost for gateway peripherals."""

    usb_error_timeout = 10

    def __init__(
        self,
        gateway_packages_queue: PackageQueueProxy,
        drivers_state_queue: multiprocessing.Queue,
        **kwargs,
    ):
        """
        Args:
            gateway_packages_queue: The Gateway package queue to be used by Drivers
                to send collected GatewayPackages.
            **kwargs: named parameters specific to this UsbDriversHost implementation.
        """
        super(GenericDriversHost, self).__init__(
            gateway_packages_queue,
            drivers_state_queue,
            self.usb_error_timeout,
            **kwargs,
        )
        log.info("Generic driver host is ready.")

    def connect(self):
        """Connects custom drivers"""
        for d in self.drivers:
            d.connect_device()

    def disconnect(self):
        """Disconnects custom drivers"""
        for d in self.drivers:
            d.disconnect_device()

    def discover_physical_ids(self) -> List[DeviceId]:
        """Returns and empty.

        Returns:
            The List empty.
        """
        log.warning("This is a generic host for gateway peripherals which does not need to be produced (I2C, SPI,etc).")
        return list()

    def get_driver_host_name(self):
        return "Generic"

    def _add_drivers(self, edges: List[EdgeDescriptor]):
        """Adds the set of drivers for the specified edges in the host.

        Args:
            edges: List of EdgeDescriptor's for the drivers to be created in
                the host instance.
        """
        for descriptor in edges:
            edge_driver_param = descriptor.edge_parameters.get('custom_driver', None)

            if not edge_driver_param:
                log.error(f"'custom_driver' parameter not specified for edge {descriptor.logical_id} "
                          f"with type: {descriptor.device_kind}")
                log.warning(f"The edge {descriptor.logical_id} could not be created.")

            driver_class = _GeneralDriversRegistry.get(edge_driver_param, None)

            if not driver_class:
                log.error(f"custom_driver '{edge_driver_param}' specified for {descriptor.logical_id}"
                          f"is unknown to the {self.__class__.__name__}")
                return

            driver = driver_class(
                descriptor, self.gateway_package_queue, self._on_driver_state_changed,
                self._completed_driver_pulses_queue
            )
            log.debug(f'Driver created with {edge_driver_param} implementation for {descriptor.logical_id}')
            self.drivers.append(driver)

    def _change_driver_state(self, physical_id: Tuple, driver_state):
        drivers = list(
            filter(lambda driver: driver.physical_id == physical_id, self.drivers)
        )
        if len(drivers) > 0:
            driver = drivers.pop()
            if driver_state == DriverState.DISCONNECTED:
                driver.stop_sampling_windows()
            driver.update_driver_state(driver_state)
            self._add_drivers_state_to_queue(
                StateReport(driver.logical_id, driver_state)
            )
        else:
            log.error("DEVICE IS NOT IN THE CURRENT GATEWAY CONFIGURATION: {}".format(physical_id))

    def take_fix_error_action(self, serial):
        pass



"""
This module declares a MockDriver that is currently used in some
unit test.
"""
# ============================ imports =========================================
from txp.common.edge import EdgeDescriptor
from txp.common.edge import DeviceId
from txp.common.edge import (
    VibrationMock, ImageMock
)
from txp.devices.package_queue import PackageQueueProxy
from txp.devices.drivers.drivers_host import DriversHost
from txp.devices.drivers.driver import (
    Driver,
    DriverState,
    StateReport,
    MalformedEdgeParameterError
)
import threading
import time
import numpy as np
from typing import List, Tuple, Set
import multiprocessing
from txp.common.config import settings
import logging
from enum import Enum
from dataclasses import dataclass
import queue
import os

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Class to define a signal component in the frequency domain
# ==============================================================================

class ComponentType(Enum):
    Sine = "sine"
    Cosine = "cosine"
    ComplexExponential = "exp"


@dataclass
class SignalComponentDefinition:
    amplitude: float  # any sematic unit
    frequency: float  # Hertz: cycles per seconds
    phase: float  # seconds
    component_type: ComponentType = ComponentType.Sine

    @property
    def angular_frequency(self) -> float:
        return self.frequency * 2.0 * np.pi

    @property
    def phase_in_radians(self) -> float:
        return self.phase * 2.0 * np.pi


default_components = [
    SignalComponentDefinition(amplitude=3, frequency=1000.0, phase=0.0),
]


# =============================================================================
# Class and functions to define and work with signals
# =============================================================================


@dataclass
class SignalComponent:
    """Spectral Component of a signal class.

    You can define a component for a signal using this class

    Args:
        sampling_frequency: frecuency of sampling.
        definition: SignalComponentDefinition instance for parameters of the spectral component.
        time_axis: Component time axis.
    """

    definition: SignalComponentDefinition
    sampling_frequency: float
    time_axis: np.ndarray

    @property
    def values(self) -> np.ndarray:
        """Given the time axis received in the constructor, we compute
        the values of the signal component."""
        if self.definition.component_type == ComponentType.Sine:
            return self.definition.amplitude * np.sin(self.definition.angular_frequency * self.time_axis)
        if self.definition.component_type == ComponentType.Cosine:
            return self.definition.amplitude * np.cos(self.definition.angular_frequency * self.time_axis)
        if self.definition.component_type == ComponentType.ComplexExponential:
            return np.exp(1.j * self.definition.angular_frequency * self.time_axis)
        else:
            raise Exception("No supported component")


class Signal:
    """Signal class.

    Defines the signal of a dimension in the mock driver

    Args:
        periods_to_synthetize: .
        sampling_frequency: Frequency of sampling.
        spectral_components: List of SignalComponentDefinition, a signal is composed of many components.
    """

    def __init__(self, periods_to_synthetize: float = 38.0, sampling_frequency: float = 800,
                 spectral_components: List[SignalComponentDefinition] = default_components) -> None:
        self.periods_to_synthetize = periods_to_synthetize
        self.sampling_frequency = sampling_frequency
        self.spectral_components = spectral_components
        self.frequencies = [sc.frequency for sc in self.spectral_components]
        self.highest_frequency = max(self.frequencies)
        if (2 * self.highest_frequency) > self.sampling_frequency:
            print("WARNING: your sampling frequency is too low!!!")
        self.lowest_frequency = min(self.frequencies)
        self.signal_period = 1.0 / self.lowest_frequency
        self.sample_count = int((self.sampling_frequency / self.lowest_frequency) * self.periods_to_synthetize)
        self.end_time = self.signal_period * self.periods_to_synthetize
        self.time_axis = np.linspace(0.0, self.end_time, self.sample_count, endpoint=False)
        self.components = [SignalComponent(component, self.sampling_frequency, self.time_axis)
                           for component in self.spectral_components]
        self.signal_amplitude_axis = sum([c.values for c in self.components])
        assert self.sample_count == len(self.signal_amplitude_axis)


def get_signals_per_dimension(edge):
    """Creates signals of each dimension specified in edge descriptor given as argument.
    Args:
        edge: edge with configuration for signals, this function is assuming only one perception.
    """
    components_per_dimension = []
    for components_per_dim in edge.perceptions[VibrationMock.perception_name()][
                                                                            MockDriver.PARAM_COMPONENTS_PER_DIMENSION]:
        components = []
        for component in components_per_dim["components"]:
            components.append(
                SignalComponentDefinition(amplitude=component["amplitude"],
                                          frequency=component["frequency"],
                                          phase=component["phase"],
                                          component_type=ComponentType(component["component_type"]))
            )
        components_per_dimension.append(components)
    signals_per_dimension = []
    for i in range(0, edge.perceptions[VibrationMock.perception_name()][MockDriver.PERCEPTION_DIMENSIONS_FIELD][0]):
        signal = Signal(sampling_frequency=edge.perceptions[VibrationMock.perception_name()][
            MockDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD],
                        spectral_components=components_per_dimension[i])
        signals_per_dimension.append(signal)
    return signals_per_dimension


def get_image_bytes(image_location):
    with open(image_location, "rb") as image_file:
        image_bytes = image_file.read()
    img = list(image_bytes)
    return (1, len(img)), [img]


# ==============================================================================
# MockDriver class used to test the state machine
# ==============================================================================
class MockDriver(Driver):
    """Mock driver class.

    Defines a known signals generator device for infrastructure testing.

    """

    PARAM_SAMPLE_ALL = "all_possible_samples"
    PARAM_COMPONENTS_PER_DIMENSION = "components_per_dimension"
    CLOCK_PULSE_PARAM = "clock_pulse"

    def __init__(
            self,
            edge_descriptor: EdgeDescriptor,
            package_queue: PackageQueueProxy,
            on_driver_state_changed: callable,
            on_driver_change_state_callback: callable,
            host_clock_notification_queue: queue.Queue
    ):
        Driver.__init__(self, edge_descriptor, package_queue, on_driver_state_changed, host_clock_notification_queue)

        self.on_driver_change_state_callback = on_driver_change_state_callback
        self.vibration_params = {}
        self.image_params = {}
        if VibrationMock.perception_name() in self.perceptions:
            vibration_descriptor = self.perceptions[VibrationMock.perception_name()]
            if len(vibration_descriptor[self.PARAM_COMPONENTS_PER_DIMENSION]) != \
                    vibration_descriptor[self.PERCEPTION_DIMENSIONS_FIELD][0]:
                error = f"""MockDriver: components per dimension list length does not equal edge num of dimensions list 
                length: {len(vibration_descriptor[self.PARAM_COMPONENTS_PER_DIMENSION])} num of dimensions: 
                {vibration_descriptor[self.PERCEPTION_DIMENSIONS_FIELD][0]}
                """
                log.error(error)
                raise MalformedEdgeParameterError(error)

            self.vibration_params["signals_per_dimension"] = get_signals_per_dimension(edge_descriptor)
            self.vibration_params["number_of_dimensions"] = vibration_descriptor[self.PERCEPTION_DIMENSIONS_FIELD][0]
            self.vibration_params["sample_count"] = vibration_descriptor[self.PERCEPTION_DIMENSIONS_FIELD][1]
            sample_count = self.vibration_params["sample_count"]
            for i, signal in enumerate(self.vibration_params["signals_per_dimension"]):
                if signal.sample_count > sample_count:
                    warning = (
                        f"MockDriver. dimension: {i} needs a minimum sampling count of {signal.sample_count} but "
                        f"the configuration is set to {sample_count}. The slower frequencies generated by "
                        f"the mock can be lost.")
                    log.warning(warning)
                elif signal.sample_count <= sample_count:
                    message = (f"MockDriver. dimension: {i} has sampling count = {signal.sample_count} but "
                               f"the configuration is set to {sample_count}")
                    log.info(message)
            self.vibration_params["sampling_frequency"] = vibration_descriptor[
                self.PERCEPTION_SAMPLING_FREQUENCY_FIELD]
            self.vibration_params["sampling_resolution"] = 16

        if ImageMock.perception_name() in self.perceptions:
            image_descriptor = self.perceptions[ImageMock.perception_name()]
            path = os.path.join("resources", image_descriptor["image"])
            file_directory = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(file_directory, path)
            self.image_params["shape"], self.image_params["image"] = get_image_bytes(path)

        self._sample_all_possible = edge_descriptor.edge_parameters.get(self.PARAM_SAMPLE_ALL, False)

        if self.CLOCK_PULSE_PARAM in self.edge_parameters:
            self._clock_pulse: int = self.edge_parameters[self.CLOCK_PULSE_PARAM]
        else:
            self._clock_pulse: int = 20

    @classmethod
    def device_name(cls) -> str:
        return "Mock"

    def observation_pulse(self) -> int:
        return self._clock_pulse

    @classmethod
    def is_virtual(cls) -> bool:
        return False

    def connect_device(self):
        log.info(f"Connecting MockDriver {self.logical_id}")
        self.on_driver_change_state_callback(
            self.logical_id, DriverState.CONNECTED
        )

    def disconnect_device(self):
        log.info(f"Disconnecting MockDriver {self.logical_id}")
        self.on_driver_change_state_callback(
            self.logical_id, DriverState.DISCONNECTED
        )

    def _start_sampling_collect(
            self
    ) -> List["Signals"]:
        signals = []

        while True:
            if self.vibration_params:
                data = []
                for signal in self.vibration_params["signals_per_dimension"]:
                    data.append(signal.signal_amplitude_axis)
                signal = VibrationMock.build_signal_from_numpy(
                    np.array(data),
                    (self.vibration_params["number_of_dimensions"], self.vibration_params["sample_count"]),
                    self.vibration_params["sampling_frequency"],
                    self.vibration_params["sampling_resolution"],
                )
                signals.append(signal)
            if self.image_params:
                signal = ImageMock.build_signal_from_lists(
                    self.image_params["image"],
                    self.image_params["shape"],
                    0,
                    0
                )
                signals.append(signal)
            if self.image_params:
                pass
            if not self._sample_all_possible:
                break   # TODO: Sample all possible not supported right now
            else:
                time.sleep(1)

        log.debug(f"{self.logical_id} will return from sampling collect")
        return signals

    def _stop_sampling_collect(self):
        pass


# ==============================================================================
# MockDriversHost is the Host for mock drivers used in unit testing.
# ==============================================================================
class MockDriversHost(DriversHost):
    # Value requested by the super class.
    mock_driver_errored_timeout_seconds = 15

    def __init__(self, gateway_packages_queue: PackageQueueProxy,
                 drivers_state_queue: multiprocessing.Queue, **kwargs):
        """
        Args:
            gateway_packages_queue: The Gateway package queue to be used by Drivers
                to send collected GatewayPackages.
            **discovered_identifiers: identifiers to mock discovered physical devices.
        """
        super(MockDriversHost, self).__init__(gateway_packages_queue, drivers_state_queue,
                                              self.mock_driver_errored_timeout_seconds, **kwargs)

    def discover_physical_ids(self) -> Set[DeviceId]:
        import random, string
        result = []
        random_string_physical_id = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))
        result.append(DeviceId("", random_string_physical_id))
        return result

    def connect(self):
        log.debug("connecting MockDriversHost SDK")
        time.sleep(1)
        log.debug("MockDriversHost SDK connected")

    def disconnect(self):
        log.debug("Disconnecting MockDriversHost SDK")
        time.sleep(1)
        for d in self.drivers:
            d.disconnect_device()
        log.debug("MockDriversHost SDK disconnected")

    def _add_drivers(self, edges: List[EdgeDescriptor]):
        for descriptor in edges:
            driver = MockDriver(descriptor, self.gateway_package_queue,
                                self._on_driver_state_changed, self._change_driver_state,
                                self._completed_driver_pulses_queue)
            self.drivers.append(driver)
            driver.connect_device()  # Connect the driver to update states queue

    def get_driver_host_name(self):
        return MockDriver.device_name()

    def take_fix_error_action(self, serial):
        pass

    def _change_driver_state(self, serial: Tuple, driver_state):
        driver = list(filter(lambda driver: driver.logical_id == serial, self.drivers)).pop()
        driver.update_driver_state(driver_state)
        self._add_drivers_state_to_queue(StateReport(driver.logical_id, driver_state))
